# server.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2009 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

"""Servers for the KAT device control language."""

from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa E402

import logging
import re
import socket
import sys
import threading
import time
import traceback

from builtins import object, range
from collections import deque
from concurrent.futures import Future
from functools import partial, wraps

import tornado.ioloop
import tornado.tcpserver

from _thread import get_ident as get_thread_ident
from future.utils import with_metaclass
from past.builtins import basestring
from tornado import gen, iostream
from tornado.concurrent import Future as tornado_Future
from tornado.concurrent import chain_future
from tornado.util import ObjectDict

# 'import katcp' so that we can use katcp.__version__ later
# we cannot do this: 'from . import __version__' because __version__
# does not exist at this stage
import katcp

from .core import (DEFAULT_KATCP_MAJOR, MS_TO_SEC_FAC, SEC_TO_MS_FAC,
                   SEC_TS_KATCP_MAJOR, VERSION_CONNECT_KATCP_MAJOR, AsyncReply,
                   DeviceServerMetaclass, FailReply, Message, MessageParser,
                   ProtocolFlags, ensure_native_str)
from .ioloop_manager import IOLoopManager, with_relative_timeout
from .kattypes import (Int, Str, has_katcp_protocol_flags,
                       minimum_katcp_version, request, return_reply)
from .sampling import (SampleNone, SampleStrategy, format_inform_v4,
                       format_inform_v5)

log = logging.getLogger("katcp.server")

BASE_REQUESTS = frozenset(['client-list',
                           'halt',
                           'help',
                           'log-level',
                           'new-command',
                           'raise-exception',
                           'raise-fail',
                           'restart',
                           'sensor-list',
                           'sensor-sampling-clear',
                           'sensor-value',
                           'version-list',
                           'watchdog',
                           'sensor-sampling', ])
"List of basic KATCP requests that a minimal device server should implement"


def return_future(fn):
    """Decorator that turns a synchronous function into one returning a future.

    This should only be applied to non-blocking functions. Will do set_result()
    with the return value, or set_exc_info() if an exception is raised.

    """
    @wraps(fn)
    def decorated(*args, **kwargs):
        return gen.maybe_future(fn(*args, **kwargs))

    return decorated


def construct_name_filter(pattern):
    """Return a function for filtering sensor names based on a pattern.

    Parameters
    ----------
    pattern : None or str
        If None, the returned function matches all names.
        If pattern starts and ends with '/' the text between the slashes
        is used as a regular expression to search the names.
        Otherwise the pattern must match the name of the sensor exactly.

    Returns
    -------
    exact : bool
        Return True if pattern is expected to match exactly. Used to
        determine whether having no matching sensors constitutes an error.
    filter_func : f(str) -> bool
        Function for determining whether a name matches the pattern.

    """
    if pattern is None:
        return False, lambda name: True
    if pattern.startswith('/') and pattern.endswith('/'):
        name_re = re.compile(pattern[1:-1])
        return False, lambda name: name_re.search(name) is not None
    return True, lambda name: name == pattern


class ClientConnection(object):
    """Encapsulates the connection between a single client and the server."""

    def __init__(self, server, conn_id):
        self._server = server
        self._conn_key = conn_id
        self._disconnect_called = False
        self._get_address = partial(server.get_address, conn_id)
        self._send_message = partial(server.send_message, conn_id)
        self._mass_send_message = server.mass_send_message
        self.flush_on_close = partial(server.flush_on_close, conn_id)

    @property
    def address(self):
        return self._get_address()

    @property
    def client_disconnect_called(self):
        return self._disconnect_called

    def disconnect(self, reason):
        """Disconnect this client connection for specified reason"""
        self._server._disconnect_client(self._conn_key, self, reason)

    def inform(self, msg):
        """Send an inform message to a particular client.

        Should only be used for asynchronous informs. Informs
        that are part of the response to a request should use
        :meth:`reply_inform` so that the message identifier
        from the original request can be attached to the
        inform.

        Parameters
        ----------
        msg : Message object
            The inform message to send.

        """
        assert (msg.mtype == Message.INFORM)
        return self._send_message(msg)

    def reply_inform(self, inform, orig_req):
        """Send an inform as part of the reply to an earlier request.

        Parameters
        ----------
        inform : Message object
            The inform message to send.
        orig_req : Message object
            The request message being replied to. The inform message's
            id is overridden with the id from orig_req before the
            inform is sent.

        """
        assert (inform.mtype == Message.INFORM)
        assert (inform.name == orig_req.name)
        inform.mid = orig_req.mid
        return self._send_message(inform)

    def mass_inform(self, msg):
        """Send an inform message to all clients.

        Parameters
        ----------
        msg : Message object
            The inform message to send.

        """
        assert (msg.mtype == Message.INFORM)
        return self._mass_send_message(msg)

    def reply(self, reply, orig_req):
        """Send an asynchronous reply to an earlier request.

        Parameters
        ----------
        reply : Message object
            The reply message to send.
        orig_req : Message object
            The request message being replied to. The reply message's
            id is overridden with the id from orig_req before the
            reply is sent.

        """
        assert (reply.mtype == Message.REPLY)
        assert reply.name == orig_req.name
        reply.mid = orig_req.mid
        return self._send_message(reply)

    def on_client_disconnect_was_called(self):
        """Prevent multiple calls to on_client_disconnect handler.

        Call this when an on_client_disconnect handler has been called.

        """
        if self._disconnect_called:
            raise RuntimeError('"on_client_disconnect" already called '
                               'for this connection')
        self._disconnect_called = True


class ThreadsafeClientConnection(ClientConnection):
    """Make ClientConnection compatible with messages sent from other threads."""
    def __init__(self, server, conn_id):
        super(ThreadsafeClientConnection, self).__init__(server, conn_id)
        self._send_message = partial(server.send_message_from_thread, conn_id)
        self._mass_send_message = server.mass_send_message_from_thread


class KATCPServer(object):
    """Tornado IO backend for a KATCP Device.

    Listens for connections on a server socket, reads KATCP messages off
    the wire and passes them on to a DeviceServer-like class.

    All class CONSTANT attributes can be changed until start() is called.

    """
    BACKLOG = 5              # Size of server socket backlog
    MAX_MSG_SIZE = 2*1024*1024
    """Maximum message size that can be received in bytes.

    If more than MAX_MSG_SIZE bytes are read from the client without
    encountering a message terminator (i.e. newline), the connection is closed.

    """
    MAX_WRITE_BUFFER_SIZE = 2*MAX_MSG_SIZE
    """Maximum outstanding bytes to be buffered by the server process.

    If more than MAX_WRITE_BUFFER_SIZE bytes are outstanding, the client
    connection is closed. Note that the OS also buffers socket writes,
    so more than MAX_WRITE_BUFFER_SIZE bytes may be untransmitted in total.

    """
    DISCONNECT_TIMEOUT = 1
    """How long to wait for the device on_client_disconnect() to complete.

    Note that this will only work if the device on_client_disconnect() method
    is non-blocking (i.e. returns a future immediately). Otherwise the ioloop
    will be blocked and unable to apply the timeout.

    """

    client_connection_factory = ClientConnection
    """Factory that produces a ClientConnection compatible instance.

    signature: client_connection_factory(server, conn_id)

    Should be set before calling start().

    """

    def __init__(self, device, host, port, tb_limit=20, logger=log):
        """Initialise the IO server instance.

        Parameters
        ----------
        device : object
            :class:`katcp.DeviceServer`-like KATCP message-eating device
        host : str
            IP to bind the server socket on
        port : int
            Port to listen on
        tb_limit : int, optional
            Maximum traceback length that is sent to clients when an unhandled
            exception is encountered
        logger : :class:`logging.Logger` object, optional
            Logger instance to use for logging, defaults to module log

        API with device
        ===============

        Device methods called
        ---------------------
        All device methods called should be non-blocking. Expects all
        device.on_* methods to return a future that resolves when it is done
        or ready to be called again.

        ready = device.on_client_connect(client_conn)

            client_conn -- ClientConnection-like instance returned by
                           self.client_connection_factory
            ready -- Future that resolves when device is ready to accept messages

        ready = device.on_client_disconnect(client_conn, reason, connection_open)

            client_conn -- as above
            reason -- string reason for disconnect
            connection_open -- True if the connection can still be written to
            ready -- Future that resolves when the connection can be closed

        ready = device.on_message(client_conn, msg)

            client_conn -- as above
            msg -- katcp.Message object representing the most recent message
            ready -- Future that resolves when device is ready to accept another message

        inform = device.create_log_inform(level_name, msg, name, timestamp=None)

            inform -- #log inform message

        Services provided to device
        ===========================

        It is expected that device will use the KATCPServer instance's ioloop.
        Devices would mainly interact using:

        set_ioloop(), start(), stop(), join(), send_message(),
        mass_send_message(), and _from_thread() versions of message sending
        methods. Except for the _from_thread() methods, all calls must be made
        from the ioloop thread.

        """
        self._device = device
        self._bindaddr = (host, port)
        self._tb_limit = tb_limit
        self._logger = logger
        self._parser = MessageParser()
        # Indicate that server is running and ready to accept connections
        self._running = threading.Event()
        # Indicate that we are stopped, i.e. join() can return
        self._stopped = threading.Event()
        self.ioloop = None
        "The Tornado IOloop to use, set by self.set_ioloop()"
        # ID of Thread that hosts the IOLoop.
        # Used to check that we are running in the ioloop.
        self.ioloop_thread_id = None
        # Map from tornado IOStreams to ClientConnection objects
        self._connections = {}
        self._ioloop_manager = IOLoopManager(managed_default=True)

    @property
    def bind_address(self):
        """The (host, port) where the server is listening for connections."""
        return self._bindaddr

    def setDaemon(self, daemonic):
        """Set daemonic state of the managed ioloop thread to True / False

        Calling this method for a non-managed ioloop has no effect. Must be called before
        start(), or it will also have no effect
        """
        self._ioloop_manager.setDaemon(daemonic)

    def set_ioloop(self, ioloop=None):
        """Set the tornado IOLoop to use.

        Sets the tornado.ioloop.IOLoop instance to use, defaulting to
        IOLoop.current(). If set_ioloop() is never called the IOLoop is started
        in a new thread, and will be stopped if self.stop() is called.

        Notes
        -----
        Must be called before start() is called.

        """
        self._ioloop_manager.set_ioloop(ioloop, managed=False)

    def start(self, timeout=None):
        """Install the server on its IOLoop, optionally starting the IOLoop.

        Parameters
        ----------
        timeout : float or None, optional
            Time in seconds to wait for server thread to start.

        """
        if self._running.isSet():
            raise RuntimeError('Server already started')
        self._stopped.clear()
        # Make sure we have an ioloop
        self.ioloop = self._ioloop_manager.get_ioloop()
        self._ioloop_manager.start()
        # Set max_buffer_size to ensure streams are closed
        # if too-large messages are received
        self._tcp_server = tornado.tcpserver.TCPServer(
            self.ioloop, max_buffer_size=self.MAX_MSG_SIZE)
        self._tcp_server.handle_stream = self._handle_stream
        self._server_sock = self._bind_socket(self._bindaddr)
        self._bindaddr = self._server_sock.getsockname()

        self.ioloop.add_callback(self._install)
        if timeout:
            return self._running.wait(timeout)

    def stop(self, timeout=1.0):
        """Stop a running server (from another thread).

        Parameters
        ----------
        timeout : float or None, optional
            Seconds to wait for server to have *started*.

        Returns
        -------
        stopped : thread-safe Future
            Resolves when the server is stopped

        """
        if timeout:
            self._running.wait(timeout)
        return self._ioloop_manager.stop(callback=self._uninstall)

    def join(self, timeout=None):
        """Rejoin the server thread.

        Parameters
        ----------
        timeout : float or None, optional
            Time in seconds to wait for the thread to finish.

        Notes
        -----
        If the ioloop is not managed, this function will block until the server
        port is closed, meaning a new server can be started on the same port.

        """
        t0 = time.time()
        self._ioloop_manager.join(timeout=timeout)
        if timeout:
            self._stopped.wait(timeout - (time.time() - t0))

    def running(self):
        """Whether the handler thread is running."""
        return self._running.isSet()

    def wait_running(self, timeout=None):
        """Wait until the handler thread is running."""
        return self._running.wait(timeout)

    def _bind_socket(self, bindaddr):
        """Create a listening server socket."""
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setblocking(0)
        try:
            sock.bind(bindaddr)
        except Exception:
            self._logger.exception("Unable to bind to %s" % str(bindaddr))
            raise
        sock.listen(self.BACKLOG)
        return sock

    def _install(self):
        # Do stuff to put us on the IOLoop
        self.ioloop_thread_id = get_thread_ident()
        self._tcp_server.add_socket(self._server_sock)
        self._running.set()

    @gen.coroutine
    def _uninstall(self):
        # Stop listening, close all open connections and remove us from IOLoop
        assert get_thread_ident() == self.ioloop_thread_id
        try:
            self._tcp_server.stop()
            for stream, conn in list(self._connections.items()):
                yield self._disconnect_client(stream, conn,
                                              'Device server shutting down.')
        finally:
            self.ioloop = None
            self._running.clear()
            self._stopped.set()

    @gen.coroutine
    def _handle_stream(self, stream, address):
        """Handle a new connection as a tornado.iostream.IOStream instance."""
        try:
            assert get_thread_ident() == self.ioloop_thread_id
            stream.set_close_callback(partial(self._stream_closed_callback,
                                              stream))
            # Our message packets are small, don't delay sending them.
            stream.set_nodelay(True)
            stream.max_write_buffer_size = self.MAX_WRITE_BUFFER_SIZE

            # Abuse IOStream object slightly by adding 'address' and 'closing'
            # attributes. Use nasty prefix to prevent naming collisions.
            stream.KATCPServer_address = address
            # Flag to indicate that no more write should be accepted so that
            # we can flush the write buffer when closing a connection
            stream.KATCPServer_closing = False

            client_conn = self.client_connection_factory(self, stream)
            self._connections[stream] = client_conn
            try:
                yield gen.maybe_future(self._device.on_client_connect(client_conn))
            except Exception:
                # If on_client_connect fails there is no reason to continue
                # trying to handle this connection. Try and send exception info
                # to the client and disconnect
                e_type, e_value, trace = sys.exc_info()
                reason = "\n".join(traceback.format_exception(
                    e_type, e_value, trace, self._tb_limit))
                log_msg = 'Device error initialising connection {0}'.format(reason)
                self._logger.error(log_msg)
                data = bytes(Message.inform('log', log_msg)) + b"\n"
                stream.write(data)
                stream.close(exc_info=True)
            else:
                self._line_read_loop(stream, client_conn)
        except Exception:
            self._logger.error('Unhandled exception trying '
                               'to handle new connection', exc_info=True)

    @gen.coroutine
    def _line_read_loop(self, stream, client_conn):
        assert get_thread_ident() == self.ioloop_thread_id
        client_address = self.get_address(stream)
        try:
            while True:
                try:
                    if stream.closed():
                        # read_until_regex() below will still give us the
                        # backlogged messages buffered by the iostream,
                        # resulting in message handlers being called with a
                        # closed connection. Exit early instead.
                        break
                    line = yield stream.read_until_regex(b'\n|\r')
                except iostream.StreamClosedError:
                    # Assume that _stream_closed_callback() will handle this
                    break
                except Exception:
                        self._logger.warn('Unhandled Exception '
                                          'while reading from client {0}:'
                                          .format(client_address), exc_info=True)
                try:
                    line = line.replace(b"\r", b"\n").split(b"\n")[0]
                    msg = self._parser.parse(line) if line else None
                except Exception:
                    msg = None
                    e_type, e_value, trace = sys.exc_info()
                    reason = "\n".join(traceback.format_exception(
                        e_type, e_value, trace, self._tb_limit))
                    self._logger.error("BAD COMMAND: %s in line %r"
                                       % (reason, line))
                    self.send_message(
                        stream, self._device.create_log_inform("error", reason, "root"))
                    continue  # Wait for the next message and hope it is better
                try:
                    if msg:  # Ignore empty messages (i.e empty lines)
                        yield self._device.on_message(client_conn, msg)
                except Exception:
                    self._logger.error('Error handling message {0!s}'
                                       .format(msg), exc_info=True)
                # Allow the ioloop to run since we may be starving it if
                # there is buffered data in the stream, resulting in yield
                # stream.read_until_regex('\n|\r') never actually yielding
                # control to the ioloop
                yield gen.moment
        except Exception:
            self._logger.error('Unexpected exception in read-loop for client {0}:'
                               .format(client_address))
        finally:
            self._logger.info('Reading loop for client {0} completed'
                              .format(client_address))

    def _stream_closed_callback(self, stream):
        assert get_thread_ident() == self.ioloop_thread_id
        # Remove ClientConnection object for the current stream from our state
        conn = self._connections.pop(stream, None)
        error_repr = '{0!r}'.format(stream.error) if stream.error else ''
        if error_repr:
            self._logger.warn('Stream for client {0} closed with error {1}'
                              .format(self.get_address(stream), error_repr))
        if conn:
            reason = error_repr or "Socket EOF"
            # Return the future from _disconnect_client()
            return self._disconnect_client(stream, conn, reason)

    @gen.coroutine
    def _disconnect_client(self, stream, conn, reason):
        assert get_thread_ident() == self.ioloop_thread_id
        stream_open = not stream.closed()
        address = self.get_address(stream)
        try:
            if not conn.client_disconnect_called:
                try:
                    conn.on_client_disconnect_was_called()
                    f = gen.maybe_future(self._device.on_client_disconnect(
                        conn, reason, stream_open))
                    yield with_relative_timeout(self.DISCONNECT_TIMEOUT, f)
                except Exception:
                    self._logger.error('Error while calling '
                                       'on_client_disconnect for client {0}'
                                       .format(address), exc_info=True)

        finally:
            # Make sure stream is closed.
            stream.close()

    def get_address(self, stream):
        """Text representation of the network address of a connection stream.

        Notes
        -----
        This method is thread-safe

        """
        try:
            addr = ":".join(str(part) for part in stream.KATCPServer_address)
        except AttributeError:
            # Something weird happened, but keep trucking
            addr = '<error>'
            self._logger.warn('Could not determine address of stream',
                              exc_info=True)
        return addr

    def send_message(self, stream, msg):
        """Send an arbitrary message to a particular client.

        Parameters
        ----------
        stream : :class:`tornado.iostream.IOStream` object
            The stream to send the message to.
        msg : Message object
            The message to send.

        Notes
        -----
        This method can only be called in the IOLoop thread.

        Failed sends disconnect the client connection and calls the device
        on_client_disconnect() method. They do not raise exceptions, but they
        are logged. Sends also fail if more than self.MAX_WRITE_BUFFER_SIZE
        bytes are queued for sending, implying that client is falling behind.

        """
        assert get_thread_ident() == self.ioloop_thread_id
        try:
            if stream.KATCPServer_closing or stream.closed():
                raise RuntimeError(
                    'Stream is closing or closed so we cannot accept any more writes')
            return stream.write(bytes(msg) + b'\n')
        except Exception:
            addr = self.get_address(stream)
            self._logger.warn('Could not send message {0!r} to {1}'
                              .format(str(msg), addr), exc_info=True)
            stream.close(exc_info=True)

    def flush_on_close(self, stream):
        """Flush tornado iostream write buffer and prevent further writes.

        Returns a future that resolves when the stream is flushed.

        """
        assert get_thread_ident() == self.ioloop_thread_id
        # Prevent futher writes
        stream.KATCPServer_closing = True
        # Write empty message to get future that resolves when buffer is flushed
        return stream.write(b'\n')

    def call_from_thread(self, fn):
        """Allow thread-safe calls to ioloop functions.

        Uses add_callback if not in the IOLoop thread, otherwise calls
        directly. Returns an already resolved `tornado.concurrent.Future` if in
        ioloop, otherwise a `concurrent.Future`. Logs unhandled exceptions.
        Resolves with an exception if one occurred.

        """
        if self.in_ioloop_thread():
            f = tornado_Future()
            try:
                f.set_result(fn())
            except Exception as e:
                f.set_exception(e)
                self._logger.exception('Error executing callback '
                                       'in ioloop thread')
            finally:
                return f
        else:
            f = Future()
            try:
                f.set_running_or_notify_cancel()

                def send_message_callback():
                    try:
                        f.set_result(fn())
                    except Exception as e:
                        f.set_exception(e)
                        self._logger.exception(
                            'Error executing wrapped async callback')

                self.ioloop.add_callback(send_message_callback)
            finally:
                return f

    def send_message_from_thread(self, stream, msg):
        """Thread-safe version of send_message() returning a Future instance.

        Returns
        -------
        A Future that will resolve without raising an exception as soon as
        the call to send_message() completes. This does not guarantee that the
        message has been delivered yet. If the call to send_message() failed,
        the exception will be logged, and the future will resolve with the
        exception raised. Since a failed call to send_message() will result
        in the connection being closed, no real error handling apart from
        logging will be possible.

        Notes
        -----
        This method is thread-safe. If called from within the ioloop,
        send_message is called directly and a resolved tornado.concurrent.Future
        is returned, otherwise a callback is submitted to the ioloop that will
        resolve a thread-safe concurrent.futures.Future instance.

        """
        return self.call_from_thread(partial(self.send_message, stream, msg))

    def mass_send_message(self, msg):
        """Send a message to all connected clients.

        Notes
        -----
        This method can only be called in the IOLoop thread.

        """
        for stream in self._connections.keys():
            if not stream.closed():
                # Don't cause noise by trying to write to already closed streams
                self.send_message(stream, msg)

    def mass_send_message_from_thread(self, msg):
        """Thread-safe version of send_message() returning a Future instance.

        See return value and notes for send_message_from_thread().

        """
        return self.call_from_thread(partial(self.mass_send_message, msg))

    def in_ioloop_thread(self):
        """Return True if called in the IOLoop thread of this server."""
        return get_thread_ident() == self.ioloop_thread_id


class ClientRequestConnection(object):
    """Encapsulates specific KATCP request and associated client connection."""

    def __init__(self, client_connection, req_msg):
        self.client_connection = client_connection
        assert(req_msg.mtype == Message.REQUEST)
        self.msg = req_msg
        self.mass_inform = client_connection.mass_inform

    def inform(self, *args):
        inf_msg = Message.reply_inform(self.msg, *args)
        return self.client_connection.inform(inf_msg)

    def reply(self, *args):
        rep_msg = Message.reply_to_request(self.msg, *args)
        self._post_reply()
        return self.client_connection.reply(rep_msg, self.msg)

    def reply_with_message(self, rep_msg):
        """Send a pre-created reply message to the client connection.

        Will check that rep_msg.name matches the bound request.

        """
        self._post_reply()
        return self.client_connection.reply(rep_msg, self.msg)

    def _post_reply(self):
        # Future calls to reply_*() and inform() should error out.
        self.reply = self.reply_again
        self.reply_with_message = self.reply_again
        self.inform = self.inform_after_reply

    def reply_again(self, *args):
        raise RuntimeError('Reply to request %r already sent.' % self.msg)

    def inform_after_reply(self, *args):
        raise RuntimeError('Cannot send inform for '
                           'request %r after sending reply.' % self.msg)

    def make_reply(self, *args):
        return Message.reply_to_request(self.msg, *args)


class MessageHandlerThread(object):
    """Provides backwards compatibility for server expecting its own thread."""
    def __init__(self, handler, log_inform_formatter, logger=log):
        self.handler = handler
        try:
            owner = handler.__self__.name
        except AttributeError:
            owner = handler.__self__.__class__.__name__
        self.name = "{}.message_handler" .format(owner)
        self.log_inform_formatter = log_inform_formatter
        self._wake = threading.Event()
        self._msg_queue = deque()
        self._logger = logger
        self._running = threading.Event()
        self._thread = None
        self.ioloop = None
        super(MessageHandlerThread, self).__init__()

    def set_ioloop(self, ioloop):
        self.ioloop = ioloop

    def on_message(self, client_conn, msg):
        """Handle message.

        Returns
        -------
        ready : Future
            A future that will resolve once we're ready, else None.

        Notes
        -----
        *on_message* should not be called again until *ready* has resolved.

        """
        MAX_QUEUE_SIZE = 300
        if len(self._msg_queue) >= MAX_QUEUE_SIZE:
            # This should never happen if callers to handle_message wait
            # for its futures to resolve before sending another message.
            # NM 2014-10-06: Except when there are multiple clients. Oops.
            raise RuntimeError('MessageHandlerThread unhandled '
                               'message queue full, not handling message')
        ready_future = Future()
        self._msg_queue.append((ready_future, client_conn, msg))
        self._wake.set()
        return ready_future

    def run(self):
        self._running.set()
        # Set the default ioloop for anything using IOLoop.current()
        # in this thread gets self.ioloop
        self.ioloop.make_current()
        try:
            while self._running.isSet():
                if not self._msg_queue:
                    self._wake.wait()
                while self._msg_queue:
                    ready_future, client_conn, msg = self._msg_queue.popleft()
                    try:
                        res = self.handler(client_conn, msg)
                        if gen.is_future(res):
                            self.ioloop.add_callback(chain_future, res,
                                                     ready_future)
                        else:
                            ready_future.set_result(res)
                    except Exception as e:
                        err_msg = ('Error calling message '
                                   'handler for msg:\n {0!s}'.format(msg))
                        self._logger.error(err_msg, exc_info=True)
                        client_conn.inform(self.log_inform_formatter(
                            'error', 'See device logs:\n' + err_msg, 'root'))
                        ready_future.set_exception(e)
                self._wake.clear()

        except Exception:
            self._logger.error(
                'Unhandled exception in message handler thread: ', exc_info=True)
        finally:
            self._running.clear()
            self._logger.info('Message handler thread stopped.')

    def start(self, timeout=None):
        if self._thread and self._thread.isAlive():
            raise RuntimeError('Cannot start since thread is already running')
        self._thread = threading.Thread(target=self.run, name=self.name)
        self._thread.start()
        if timeout:
            return self.wait_running(timeout)

    def stop(self, timeout=1.0):
        """Stop the handler thread (from another thread).

        Parameters
        ----------
        timeout : float, optional
            Seconds to wait for server to have *started*.

        """
        if timeout:
            self._running.wait(timeout)
        self._running.clear()
        # Make sure to wake the run thread.
        self._wake.set()

    def join(self, timeout=None):
        """Rejoin the handler thread.

        Parameters
        ----------
        timeout : float or None, optional
            Time in seconds to wait for the thread to finish.

        """
        self._thread.join(timeout)

    def isAlive(self):
        return self._thread and self._thread.isAlive()

    def running(self):
        """Whether the handler thread is running."""
        return self._running.isSet()

    def wait_running(self, timeout=None):
        """Wait until the handler thread is running."""
        return self._running.wait(timeout)


class DeviceServerBase(with_metaclass(DeviceServerMetaclass, object)):
    """Base class for device servers.

    Subclasses should add .request\_* methods for dealing
    with request messages. These methods each take the client
    request connection and msg objects as arguments and should return
    the reply message or raise an exception as a result.

    Subclasses can also add .inform\_* and reply\_* methods to handle
    those types of messages.

    Should a subclass need to generate inform messages it should
    do so using either the .inform() or .mass_inform() methods.

    Finally, this class should probably not be subclassed directly
    but rather via subclassing DeviceServer itself which implements
    common .request\_* methods.

    Parameters
    ----------
    host : str
        Host to listen on.
    port : int
        Port to listen on.
    tb_limit : int, optional
        Maximum number of stack frames to send in error tracebacks.
    logger : logging.Logger object, optional
        Logger to log messages to.

    """

    ## @brief Protocol versions and flags. Default to version 5, subclasses
    ## should override PROTOCOL_INFO
    PROTOCOL_INFO = ProtocolFlags(DEFAULT_KATCP_MAJOR, 0, set([
        ProtocolFlags.MULTI_CLIENT,
        ProtocolFlags.MESSAGE_IDS,
        ]))

    def __init__(self, host, port, tb_limit=20, logger=log):
        self._server = KATCPServer(self, host, port, tb_limit, logger)
        self._logger = logger
        self._tb_limit = tb_limit
        # Thread that will optionally be used to handle requests
        self._handler_thread = None
        # Set default concurrency options
        self.set_concurrency_options()

    @property
    def bind_address(self):
        return self._server.bind_address

    def setDaemon(self, daemonic):
        """Set daemonic state of the managed ioloop thread to True / False

        Calling this method for a non-managed ioloop has no effect. Must be called before
        start(), or it will also have no effect
        """
        self._server.setDaemon(daemonic)

    def create_log_inform(self, level_name, msg, name, timestamp=None):
        """Create a katcp logging inform message.

        Usually this will be called from inside a DeviceLogger object,
        but it is also used by the methods in this class when errors
        need to be reported to the client.

        """
        if timestamp is None:
            timestamp = time.time()

        katcp_version = self.PROTOCOL_INFO.major
        timestamp_msg = ('%.6f' % timestamp
                         if katcp_version >= SEC_TS_KATCP_MAJOR
                         else str(int(timestamp*1000)))
        return Message.inform("log", level_name, timestamp_msg, name, msg)

    def on_message(self, client_conn, msg):
        """Dummy implementation of on_message required by KATCPServer.

        Will be replaced by a handler with the appropriate concurrency
        semantics when set_concurrency_options is called
        (defaults are set in __init__()).

        """
        raise RuntimeError('set_concurrency_options() '
                           'need to be called once before on_message')

    def handle_message(self, client_conn, msg):
        """Handle messages of all types from clients.

        Parameters
        ----------
        client_conn : ClientConnection object
            The client connection the message was from.
        msg : Message object
            The message to process.

        """
        # log messages received so that no one else has to
        self._logger.debug('received: {0!s}'.format(msg))

        if msg.mtype == msg.REQUEST:
            return self.handle_request(client_conn, msg)
        elif msg.mtype == msg.INFORM:
            return self.handle_inform(client_conn, msg)
        elif msg.mtype == msg.REPLY:
            return self.handle_reply(client_conn, msg)
        else:
            reason = "Unexpected message type received by server ['%s']." \
                     % (msg,)
            client_conn.inform(self.create_log_inform("error", reason, "root"))

    def handle_request(self, connection, msg):
        """Dispatch a request message to the appropriate method.

        Parameters
        ----------
        connection : ClientConnection object
            The client connection the message was from.
        msg : Message object
            The request message to process.

        Returns
        -------
        done_future : Future or None
            Returns Future for async request handlers that will resolve when
            done, or None for sync request handlers once they have completed.

        """
        send_reply = True
        # TODO Should check presence of Message-ids against protocol flags and
        # raise an error as needed.
        if msg.name in self._request_handlers:
            req_conn = ClientRequestConnection(connection, msg)
            handler = self._request_handlers[msg.name]
            try:
                reply = handler(self, req_conn, msg)
                # If we get a future, assume this is an async message handler
                # that will resolve the future with the reply message when it
                # is complete. Attach a message-sending callback to the future,
                # and return the future.
                if gen.is_future(reply):
                    concurrent = getattr(handler, '_concurrent_reply', False)
                    concurrent_str = ' CONCURRENT' if concurrent else ''

                    done_future = Future()
                    def async_reply(f):
                        try:
                            connection.reply(f.result(), msg)
                            self._logger.debug("%s FUTURE%s replied",
                                               msg.name, concurrent_str)
                        except FailReply as e:
                            reason = str(e)
                            self._logger.error("Request %s FUTURE%s FAIL: %s",
                                               msg.name, concurrent_str, reason)
                            reply = Message.reply(msg.name, "fail", reason)
                            connection.reply(reply, msg)
                        except AsyncReply:
                            self._logger.debug("%s FUTURE ASYNC OK"
                                               % (msg.name,))
                        except Exception:
                            error_reply = self.create_exception_reply_and_log(
                                msg, sys.exc_info())
                            connection.reply(error_reply, msg)
                        finally:
                            done_future.set_result(None)

                    # TODO When using the return_reply() decorator the future
                    # returned is not currently thread-safe, must either deal
                    # with it here, or in kattypes.py. Would be nice if we don't
                    # have to always fall back to adding a callback, or wrapping
                    # a thread-safe future. Supporting sync-with-thread and
                    # async futures is turning out to be a pain in the ass ;)
                    self.ioloop.add_callback(reply.add_done_callback, async_reply)
                    # reply.add_done_callback(async_reply)

                    if concurrent:
                        # Return immediately if this is a concurrent handler
                        self._logger.debug("%s FUTURE CONCURRENT OK", msg.name)
                        return
                    else:
                        self._logger.debug("%s FUTURE OK", msg.name)
                        return done_future
                else:
                    assert (reply.mtype == Message.REPLY)
                    assert (reply.name == msg.name)
                    self._logger.debug("%s OK" % (msg.name,))
            except AsyncReply as e:
                self._logger.debug("%s ASYNC OK" % (msg.name,))
                send_reply = False
            except FailReply as e:
                reason = str(e)
                self._logger.error("Request %s FAIL: %s" % (msg.name, reason))
                reply = Message.reply(msg.name, "fail", reason)
            except Exception:
                reply = self.create_exception_reply_and_log(msg, sys.exc_info())
        else:
            self._logger.error("%s INVALID: Unknown request." % (msg.name,))
            reply = Message.reply(msg.name, "invalid", "Unknown request.")

        if send_reply:
            connection.reply(reply, msg)

    def create_exception_reply_and_log(self, req_msg, exc_info):
        e_type, e_value, trace = exc_info
        reason = "\n".join(traceback.format_exception(
            e_type, e_value, trace, self._tb_limit))
        self._logger.error("Request %s FAIL: %s" % (req_msg.name, reason))
        return Message.reply(req_msg.name, "fail", reason)

    def handle_inform(self, connection, msg):
        """Dispatch an inform message to the appropriate method.

        Parameters
        ----------
        connection : ClientConnection object
            The client connection the message was from.
        msg : Message object
            The inform message to process.

        """
        if msg.name in self._inform_handlers:
            try:
                self._inform_handlers[msg.name](self, connection, msg)
            except Exception:
                e_type, e_value, trace = sys.exc_info()
                reason = "\n".join(traceback.format_exception(
                    e_type, e_value, trace, self._tb_limit))
                self._logger.error("Inform %s FAIL: %s" % (msg.name, reason))
        else:
            self._logger.warn("%s INVALID: Unknown inform." % (msg.name,))

    def handle_reply(self, connection, msg):
        """Dispatch a reply message to the appropriate method.

        Parameters
        ----------
        connection : ClientConnection object
            The client connection the message was from.
        msg : Message object
            The reply message to process.

        """
        if msg.name in self._reply_handlers:
            try:
                self._reply_handlers[msg.name](self, connection, msg)
            except Exception:
                e_type, e_value, trace = sys.exc_info()
                reason = "\n".join(traceback.format_exception(
                    e_type, e_value, trace, self._tb_limit))
                self._logger.error("Reply %s FAIL: %s" % (msg.name, reason))
        else:
            self._logger.warn("%s INVALID: Unknown reply." % (msg.name,))

    def inform(self, connection, msg):
        """Send an inform message to a particular client.

        Should only be used for asynchronous informs. Informs
        that are part of the response to a request should use
        :meth:`reply_inform` so that the message identifier
        from the original request can be attached to the
        inform.

        Parameters
        ----------
        connection : ClientConnection object
            The client to send the message to.
        msg : Message object
            The inform message to send.

        """
        if isinstance(connection, ClientRequestConnection):
            self._logger.warn(
                'Deprecation warning: do not use self.inform() '
                'within a reply handler context -- use req.inform()\n'
                'Traceback:\n %s', "".join(traceback.format_stack()))
            # Get the underlying ClientConnection instance
            connection = connection.client_connection
        connection.inform(msg)

    def mass_inform(self, msg):
        """Send an inform message to all clients.

        Parameters
        ----------
        msg : Message object
            The inform message to send.

        """
        assert (msg.mtype == Message.INFORM)
        self._server.mass_send_message_from_thread(msg)

    def reply(self, connection, reply, orig_req):
        """Send an asynchronous reply to an earlier request.

        Parameters
        ----------
        connection : ClientConnection object
            The client to send the reply to.
        reply : Message object
            The reply message to send.
        orig_req : Message object
            The request message being replied to. The reply message's
            id is overridden with the id from orig_req before the
            reply is sent.

        """
        if isinstance(connection, ClientRequestConnection):
            self._logger.warn(
                'Deprecation warning: do not use self.reply() '
                'within a reply handler context -- use req.reply(*msg_args)\n'
                'or req.reply_with_message(msg) Traceback:\n %s',
                "".join(traceback.format_stack()))
            # Get the underlying ClientConnection instance
            connection = connection.client_connection
        connection.reply(reply, orig_req)

    def reply_inform(self, connection, inform, orig_req):
        """Send an inform as part of the reply to an earlier request.

        Parameters
        ----------
        connection : ClientConnection object
            The client to send the inform to.
        inform : Message object
            The inform message to send.
        orig_req : Message object
            The request message being replied to. The inform message's
            id is overridden with the id from orig_req before the
            inform is sent.

        """
        if isinstance(connection, ClientRequestConnection):
            self._logger.warn(
                'Deprecation warning: do not use self.reply_inform() '
                'within a reply handler context -- '
                'use req.inform(*inform_arguments)\n'
                'Traceback:\n %s', "".join(traceback.format_stack()))
            # Get the underlying ClientConnection instance
            connection = connection.client_connection
        connection.reply_inform(inform, orig_req)

    def set_ioloop(self, ioloop=None):
        """Set the tornado IOLoop to use.

        Sets the tornado.ioloop.IOLoop instance to use, defaulting to
        IOLoop.current(). If set_ioloop() is never called the IOLoop is
        started in a new thread, and will be stopped if self.stop() is called.

        Notes
        -----
        Must be called before start() is called.

        """
        self._server.set_ioloop(ioloop)
        self.ioloop = self._server.ioloop

    def set_concurrency_options(self, thread_safe=True, handler_thread=True):
        """Set concurrency options for this device server.
        Must be called before :meth:`start`.

        Parameters
        ==========

        thread_safe : bool
            If True, make the server public methods thread safe. Incurs
            performance overhead.
        handler_thread : bool
            Can only be set if `thread_safe` is True. Handle all requests (even
            from different clients) in a separate, single, request-handling
            thread. Blocking request handlers will prevent the server from
            handling new requests from any client, but sensor strategies should
            still function. This more or less mimics the behaviour of a server
            in library versions before 0.6.0.

        """
        if handler_thread:
            assert thread_safe, "handler_thread=True requires thread_safe=True"
        self._server.client_connection_factory = (
            ThreadsafeClientConnection if thread_safe else ClientConnection)
        if handler_thread:
            self._handler_thread = MessageHandlerThread(
                self.handle_message, self.create_log_inform, self._logger)
            self.on_message = self._handler_thread.on_message
        else:
            self.on_message = return_future(self.handle_message)
            self._handler_thread = None

        self._concurrency_options = ObjectDict(
            thread_safe=thread_safe, handler_thread=handler_thread)

    def start(self, timeout=None):
        """Start the server in a new thread.

        Parameters
        ----------
        timeout : float or None, optional
            Time in seconds to wait for server thread to start.

        """
        if self._handler_thread and self._handler_thread.isAlive():
            raise RuntimeError('Message handler thread already started')
        self._server.start(timeout)
        self.ioloop = self._server.ioloop
        if self._handler_thread:
            self._handler_thread.set_ioloop(self.ioloop)
            self._handler_thread.start(timeout)

    def join(self, timeout=None):
        """Rejoin the server thread.

        Parameters
        ----------
        timeout : float or None, optional
            Time in seconds to wait for the thread to finish.

        """
        self._server.join(timeout)
        if self._handler_thread:
            self._handler_thread.join(timeout)

    def stop(self, timeout=1.0):
        """Stop a running server (from another thread).

        Parameters
        ----------
        timeout : float, optional
            Seconds to wait for server to have *started*.

        Returns
        -------
        stopped : thread-safe Future
            Resolves when the server is stopped

        """
        stopped = self._server.stop(timeout)
        if self._handler_thread:
            self._handler_thread.stop(timeout)

        return stopped

    def running(self):
        """Whether the server is running."""
        return self._server.running()

    def wait_running(self, timeout=None):
        """Wait until the server is running"""
        return self._server.wait_running(timeout)

    @return_future
    def on_client_connect(self, conn):
        """Called after client connection is established.

        Subclasses should override if they wish to send clients
        message or perform house-keeping at this point.

        Parameters
        ----------
        conn : ClientConnection object
            The client connection that has been successfully established.

        Returns
        -------
        Future that resolves when the device is ready to accept messages.

        """
        pass

    @return_future
    def on_client_disconnect(self, conn, msg, connection_valid):
        """Called before a client connection is closed.

        Subclasses should override if they wish to send clients
        message or perform house-keeping at this point. The server
        cannot guarantee this will be called (for example, the client
        might drop the connection). The message parameter contains
        the reason for the disconnection.

        Parameters
        ----------
        conn : ClientConnection object
            Client connection being disconnected.
        msg : str
            Reason client is being disconnected.
        connection_valid : boolean
            True if connection is still open for sending,
            False otherwise.

        Returns
        -------
        Future that resolves when the client connection can be closed.

        """
        return None

    def sync_with_ioloop(self, timeout=None):
        """Block for ioloop to complete a loop if called from another thread.

        Returns a future if called from inside the ioloop.

        Raises concurrent.futures.TimeoutError if timed out while blocking.

        """
        in_ioloop = self._server.in_ioloop_thread()
        if in_ioloop:
            f = tornado_Future()
        else:
            f = Future()

        def cb():
            f.set_result(None)

        self.ioloop.add_callback(cb)

        if in_ioloop:
            return f
        else:
            f.result(timeout)


class DeviceServer(DeviceServerBase):
    """Implements some standard messages on top of DeviceServerBase.

    Inform messages handled are:

      * version (sent on connect)
      * build-state (sent on connect)
      * log (via self.log.warn(...), etc)
      * disconnect
      * client-connected

    Requests handled are:

      * halt
      * help
      * log-level
      * restart [#restartf1]_
      * client-list
      * sensor-list
      * sensor-sampling
      * sensor-value
      * watchdog
      * version-list (only standard in KATCP v5 or later)
      * request-timeout-hint (pre-standard only if protocol flags indicates
                              timeout hints, supported for KATCP v5.1 or later)
      * sensor-sampling-clear (non-standard)

    .. [#restartf1] Restart relies on .set_restart_queue() being used to
      register a restart queue with the device. When the device needs to be
      restarted, it will be added to the restart queue.  The queue should
      be a Python Queue.Queue object without a maximum size.

    Unhandled standard requests are:

      * configure
      * mode

    Subclasses can define the tuple VERSION_INFO to set the interface
    name, major and minor version numbers. The BUILD_INFO tuple can
    be defined to give a string describing a particular interface
    instance and may have a fourth element containing additional
    version information (e.g. rc1).

    Subclasses may manipulate the versions returned by the ?version-list
    command by editing .extra_versions which is a dictionary mapping
    role or component names to (version, build_state_or_serial_no) tuples.
    The build_state_or_serial_no may be None.

    Subclasses must override the .setup_sensors() method. If they
    have no sensors to register, the method should just be a pass.

    """
    # DeviceServer has a lot of methods because there is a method
    # per request type and it's an abstract class which is only
    # used outside this module
    # pylint: disable-msg = R0904

    ## @brief Interface version information.
    VERSION_INFO = ("device_stub", 0, 1)

    ## @brief Device server build / instance information.
    BUILD_INFO = ("name", 0, 1, "")

    UNSUPPORTED_REQUESTS_BY_MAJOR_VERSION = {
        4: set(['version-list']),
    }

    SUPPORTED_PROTOCOL_MAJOR_VERSIONS = (4, 5)

    # * and ** magic fine here
    # pylint: disable-msg = W0142

    def __init__(self, *args, **kwargs):
        if self.PROTOCOL_INFO.major not in self.SUPPORTED_PROTOCOL_MAJOR_VERSIONS:
            raise ValueError('Device server only supports katcp procotol '
                             'versions %r, not %r as specified in self.PROTOCOL_INFO'
                             % (self.SUPPORTED_PROTOCOL_MAJOR_VERSIONS,
                                self.PROTOCOL_INFO.major))
        super(DeviceServer, self).__init__(*args, **kwargs)
        self.log = DeviceLogger(self, python_logger=self._logger)
        # map names to (version, build state/serial no.) tuples.
        #   None may used to indicate no build state or serial no.
        self.extra_versions = {}
        self._restart_queue = None
        self._sensors = {}  # map names to sensor objects
        # map client sockets to map of sensors -> sampling strategies
        self._strategies = {}
        # For holding ClientConnection* instances of active connections
        self._client_conns = set()

        self.setup_sensors()

    # pylint: enable-msg = W0142

    @return_future
    def on_client_connect(self, client_conn):
        """Inform client of build state and version on connect.

        Parameters
        ----------
        client_conn : ClientConnection object
            The client connection that has been successfully established.

        Returns
        -------
        Future that resolves when the device is ready to accept messages.

        """
        assert get_thread_ident() == self._server.ioloop_thread_id
        self._client_conns.add(client_conn)
        self._strategies[client_conn] = {}  # map sensors -> sampling strategies

        katcp_version = self.PROTOCOL_INFO.major
        if katcp_version >= VERSION_CONNECT_KATCP_MAJOR:
            client_conn.inform(Message.inform(
                "version-connect", "katcp-protocol", self.PROTOCOL_INFO))
            client_conn.inform(Message.inform(
                "version-connect", "katcp-library",
                "katcp-python-%s" % katcp.__version__))
            client_conn.inform(Message.inform(
                "version-connect", "katcp-device",
                self.version(), self.build_state()))
        else:
            client_conn.inform(Message.inform("version", self.version()))
            client_conn.inform(Message.inform("build-state", self.build_state()))

    def clear_strategies(self, client_conn, remove_client=False):
        """Clear the sensor strategies of a client connection.

        Parameters
        ----------
        client_connection : ClientConnection instance
            The connection that should have its sampling strategies cleared
        remove_client : bool, optional
            Remove the client connection from the strategies data-structure.
            Useful for clients that disconnect.

        """
        assert get_thread_ident() == self._server.ioloop_thread_id

        getter = (self._strategies.pop if remove_client
                  else self._strategies.get)
        strategies = getter(client_conn, None)
        if strategies is not None:
            for sensor, strategy in list(strategies.items()):
                strategy.cancel()
                del strategies[sensor]

    def on_client_disconnect(self, client_conn, msg, connection_valid):
        """Inform client it is about to be disconnected.

        Parameters
        ----------
        client_conn : ClientConnection object
            The client connection being disconnected.
        msg : str
            Reason client is being disconnected.
        connection_valid : bool
            True if connection is still open for sending,
            False otherwise.

        Returns
        -------
        Future that resolves when the client connection can be closed.

        """
        f = tornado_Future()
        @gen.coroutine
        def remove_strategies():
            self.clear_strategies(client_conn, remove_client=True)
            if connection_valid:
                client_conn.inform(Message.inform("disconnect", msg))
                yield client_conn.flush_on_close()

        try:
            self._client_conns.remove(client_conn)
            self.ioloop.add_callback(lambda: chain_future(remove_strategies(), f))
        except Exception:
            f.set_exc_info(sys.exc_info())
        return f

    def build_state(self):
        """Return build state string of the form name-major.minor[(a|b|rc)n]."""
        return "%s-%s.%s%s" % self.BUILD_INFO

    def version(self):
        """Return a version string of the form type-major.minor."""
        return "%s-%s.%s" % self.VERSION_INFO

    def add_sensor(self, sensor):
        """Add a sensor to the device.

        Usually called inside .setup_sensors() but may be called from
        elsewhere.

        Parameters
        ----------
        sensor : Sensor object
            The sensor object to register with the device server.

        """
        self._sensors[sensor.name] = sensor

    def has_sensor(self, sensor_name):
        """Whether the sensor with specified name is known."""
        return sensor_name in self._sensors

    def remove_sensor(self, sensor):
        """Remove a sensor from the device.

        Also deregisters all clients observing the sensor.

        Parameters
        ----------
        sensor : Sensor object or name string
            The sensor to remove from the device server.

        """
        if isinstance(sensor, basestring):
            sensor_name = sensor
        else:
            sensor_name = sensor.name
        sensor = self._sensors.pop(sensor_name)

        def cancel_sensor_strategies():
            for conn_strategies in self._strategies.values():
                strategy = conn_strategies.pop(sensor, None)
                if strategy:
                    strategy.cancel()
        self.ioloop.add_callback(cancel_sensor_strategies)

    def get_sensor(self, sensor_name):
        """Fetch the sensor with the given name.

        Parameters
        ----------
        sensor_name : str
            Name of the sensor to retrieve.

        Returns
        -------
        sensor : Sensor object
            The sensor with the given name.

        """
        sensor = self._sensors.get(sensor_name, None)
        if not sensor:
            raise ValueError("Unknown sensor '%s'." % (sensor_name,))
        return sensor

    def get_sensors(self):
        """Fetch a list of all sensors.

        Returns
        -------
        sensors : list of Sensor objects
            The list of sensors registered with the device server.

        """
        return list(self._sensors.values())

    def set_restart_queue(self, restart_queue):
        """Set the restart queue.

        When the device server should be restarted, it will be added to the
        queue.

        Parameters
        ----------
        restart_queue : Queue.Queue object
            The queue to add the device server to when it should be restarted.

        """
        self._restart_queue = restart_queue

    def setup_sensors(self):
        """Populate the dictionary of sensors.

        Unimplemented by default -- subclasses should add their sensors
        here or pass if there are no sensors.

        Examples
        --------
        >>> class MyDevice(DeviceServer):
        ...     def setup_sensors(self):
        ...         self.add_sensor(Sensor(...))
        ...         self.add_sensor(Sensor(...))
        ...

        """
        raise NotImplementedError("Device server subclasses must implement "
                                  "setup_sensors.")

    # request implementations

    # all requests take req and msg arguments regardless of whether
    # they're used
    # pylint: disable-msg = W0613

    def request_halt(self, req, msg):
        """Halt the device server.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether scheduling the halt succeeded.

        Examples
        --------
        ::

            ?halt
            !halt ok

        """
        f = Future()
        @gen.coroutine
        def _halt():
            req.reply("ok")
            yield gen.moment
            self.stop(timeout=None)
            raise AsyncReply

        self.ioloop.add_callback(lambda: chain_future(_halt(), f))
        return f

    def request_help(self, req, msg):
        """Return help on the available requests.

        Return a description of the available requests using a sequence of
        #help informs.

        Parameters
        ----------
        request : str, optional
            The name of the request to return help for (the default is to
            return help for all requests).

        Informs
        -------
        request : str
            The name of a request.
        description : str
            Documentation for the named request.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether sending the help succeeded.
        informs : int
            Number of #help inform messages sent.

        Examples
        --------
        ::

            ?help
            #help halt ...description...
            #help help ...description...
            ...
            !help ok 5

            ?help halt
            #help halt ...description...
            !help ok 1

        """
        if not msg.arguments:
            for name, method in sorted(self._request_handlers.items()):
                doc = method.__doc__
                req.inform(name, doc)
            num_methods = len(self._request_handlers)
            return req.make_reply("ok", str(num_methods))
        else:
            name = ensure_native_str(msg.arguments[0])
            if name in self._request_handlers:
                method = self._request_handlers[name]
                doc = method.__doc__.strip()
                req.inform(name, doc)
                return req.make_reply("ok", "1")
            return req.make_reply("fail", "Unknown request method.")

    @request(Str(optional=True))
    @return_reply(Int())
    @has_katcp_protocol_flags([ProtocolFlags.REQUEST_TIMEOUT_HINTS])
    def request_request_timeout_hint(self, req, request):
        """Return timeout hints for requests

        KATCP requests should generally take less than 5s to complete, but some
        requests are unavoidably slow. This results in spurious client timeout
        errors. This request provides timeout hints that clients can use to
        select suitable request timeouts.

        Parameters
        ----------
        request : str, optional
            The name of the request to return a timeout hint for (the default is
            to return hints for all requests that have timeout hints). Returns
            one inform per request. Must be an existing request if specified.

        Informs
        -------
        request : str
            The name of the request.
        suggested_timeout : float
            Suggested request timeout in seconds for the request. If
            `suggested_timeout` is zero (0), no timeout hint is available.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether sending the help succeeded.
        informs : int
            Number of #request-timeout-hint inform messages sent.

        Examples
        --------
        ::

            ?request-timeout-hint
            #request-timeout-hint halt 5
            #request-timeout-hint very-slow-request 500
            ...
            !request-timeout-hint ok 5

            ?request-timeout-hint moderately-slow-request
            #request-timeout-hint moderately-slow-request 20
            !request-timeout-hint ok 1

        Notes
        -----

        ?request-timeout-hint without a parameter will only return informs for
        requests that have specific timeout hints, so it will most probably be a
        subset of all the requests, or even no informs at all.

        """
        timeout_hints = {}
        if request:
            request = ensure_native_str(request)
            if request not in self._request_handlers:
                raise FailReply('Unknown request method')
            timeout_hint = getattr(
                self._request_handlers[request], 'request_timeout_hint', None)
            timeout_hint = timeout_hint or 0
            timeout_hints[request] = timeout_hint
        else:
            for request_, handler in self._request_handlers.items():
                timeout_hint = getattr(handler, 'request_timeout_hint', None)
                if timeout_hint:
                    timeout_hints[request_] = timeout_hint

        cnt = len(timeout_hints)
        for request_name, timeout_hint in sorted(timeout_hints.items()):
            req.inform(request_name, float(timeout_hint))

        return ('ok', cnt)

    def request_log_level(self, req, msg):
        """Query or set the current logging level.

        Parameters
        ----------
        level : {'all', 'trace', 'debug', 'info', 'warn', 'error', 'fatal', \
                 'off'}, optional
            Name of the logging level to set the device server to (the default
            is to leave the log level unchanged).

        Returns
        -------
        success : {'ok', 'fail'}
            Whether the request succeeded.
        level : {'all', 'trace', 'debug', 'info', 'warn', 'error', 'fatal', \
                 'off'}
            The log level after processing the request.

        Examples
        --------
        ::

            ?log-level
            !log-level ok warn

            ?log-level info
            !log-level ok info

        """
        if msg.arguments:
            try:
                self.log.set_log_level_by_name(msg.arguments[0])
            except ValueError as e:
                raise FailReply(str(e))
        return req.make_reply("ok", self.log.level_name())

    def request_restart(self, req, msg):
        """Restart the device server.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether scheduling the restart succeeded.

        Examples
        --------
        ::

            ?restart
            !restart ok

        """
        if self._restart_queue is None:
            raise FailReply("No restart queue registered -- cannot restart.")

        f = tornado_Future()

        @gen.coroutine
        def _restart():
            # .put should never block because queue should have no size limit
            self._restart_queue.put_nowait(self)
            req.reply('ok')
            raise AsyncReply

        self.ioloop.add_callback(lambda: chain_future(_restart(), f))
        return f

    def request_client_list(self, req, msg):
        """Request the list of connected clients.

        The list of clients is sent as a sequence of #client-list informs.

        Informs
        -------
        addr : str
            The address of the client as host:port with host in dotted quad
            notation. If the address of the client could not be determined
            (because, for example, the client disconnected suddenly) then
            a unique string representing the client is sent instead.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether sending the client list succeeded.
        informs : int
            Number of #client-list inform messages sent.

        Examples
        --------
        ::

            ?client-list
            #client-list 127.0.0.1:53600
            !client-list ok 1

        """
        # TODO Get list of ClientConnection* instances and implement a standard
        # 'address-print' method in the ClientConnection class
        clients = self._client_conns
        num_clients = len(clients)
        for conn in clients:
            addr = conn.address
            req.inform(addr)
        return req.make_reply('ok', str(num_clients))


    @minimum_katcp_version(5, 0)
    def request_version_list(self, req, msg):
        """Request the list of versions of roles and subcomponents.

        Informs
        -------
        name : str
            Name of the role or component.
        version : str
            A string identifying the version of the component. Individual
            components may define the structure of this argument as they
            choose. In the absence of other information clients should
            treat it as an opaque string.
        build_state_or_serial_number : str
            A unique identifier for a particular instance of a component.
            This should change whenever the component is replaced or updated.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether sending the version list succeeded.
        informs : int
            Number of #version-list inform messages sent.

        Examples
        --------
        ::

            ?version-list
            #version-list katcp-protocol 5.0-MI
            #version-list katcp-library katcp-python-0.4 katcp-python-0.4.1-py2
            #version-list katcp-device foodevice-1.0 foodevice-1.0.0rc1
            !version-list ok 3

        """
        versions = [
            ("katcp-protocol", (self.PROTOCOL_INFO, None)),
            ("katcp-library", ("katcp-python-%s" % katcp.__version__, katcp.__version__)),
            ("katcp-device", (self.version(), self.build_state())),
        ]
        extra_versions = sorted(self.extra_versions.items())

        for name, (version, build_state) in versions + extra_versions:
            if build_state is None:
                inform_args = (name, version)
            else:
                inform_args = (name, version, build_state)
            req.inform(*inform_args)

        num_versions = len(versions) + len(extra_versions)
        return req.make_reply("ok", str(num_versions))

    def request_sensor_list(self, req, msg):
        r"""Request the list of sensors.

        The list of sensors is sent as a sequence of #sensor-list informs.

        Parameters
        ----------
        name : str, optional
            Name of the sensor to list (the default is to list all sensors).
            If name starts and ends with '/' it is treated as a regular
            expression and all sensors whose names contain the regular
            expression are returned.

        Informs
        -------
        name : str
            The name of the sensor being described.
        description : str
            Description of the named sensor.
        units : str
            Units for the value of the named sensor.
        type : str
            Type of the named sensor.
        params : list of str, optional
            Additional sensor parameters (type dependent). For integer and
            float sensors the additional parameters are the minimum and maximum
            sensor value. For discrete sensors the additional parameters are
            the allowed values. For all other types no additional parameters
            are sent.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether sending the sensor list succeeded.
        informs : int
            Number of #sensor-list inform messages sent.

        Examples
        --------
        ::

            ?sensor-list
            #sensor-list psu.voltage PSU\_voltage. V float 0.0 5.0
            #sensor-list cpu.status CPU\_status. \@ discrete on off error
            ...
            !sensor-list ok 5

            ?sensor-list cpu.power.on
            #sensor-list cpu.power.on Whether\_CPU\_hase\_power. \@ boolean
            !sensor-list ok 1

            ?sensor-list /voltage/
            #sensor-list psu.voltage PSU\_voltage. V float 0.0 5.0
            #sensor-list cpu.voltage CPU\_voltage. V float 0.0 3.0
            !sensor-list ok 2

        """
        exact, name_filter = construct_name_filter(ensure_native_str(msg.arguments[0])
                                                   if msg.arguments else None)
        sensors = [(name, sensor) for name, sensor in
                   sorted(self._sensors.items()) if name_filter(name)]

        if exact and not sensors:
            return req.make_reply("fail", "Unknown sensor name.")

        self._send_sensor_value_informs(req, sensors)
        return req.make_reply("ok", str(len(sensors)))

    def _send_sensor_value_informs(self, req, sensors):
        for name, sensor in sensors:
            req.inform(name, sensor.description, sensor.units, sensor.stype,
                       *sensor.formatted_params)

    def request_sensor_value(self, req, msg):
        """Request the value of a sensor or sensors.

        A list of sensor values as a sequence of #sensor-value informs.

        Parameters
        ----------
        name : str, optional
            Name of the sensor to poll (the default is to send values for all
            sensors). If name starts and ends with '/' it is treated as a
            regular expression and all sensors whose names contain the regular
            expression are returned.

        Informs
        -------
        timestamp : float
            Timestamp of the sensor reading in seconds since the Unix
            epoch, or milliseconds for katcp versions <= 4.
        count : {1}
            Number of sensors described in this #sensor-value inform. Will
            always be one. It exists to keep this inform compatible with
            #sensor-status.
        name : str
            Name of the sensor whose value is being reported.
        value : object
            Value of the named sensor. Type depends on the type of the sensor.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether sending the list of values succeeded.
        informs : int
            Number of #sensor-value inform messages sent.

        Examples
        --------
        ::

            ?sensor-value
            #sensor-value 1244631611.415231 1 psu.voltage nominal 4.5
            #sensor-value 1244631611.415200 1 cpu.status nominal off
            ...
            !sensor-value ok 5

            ?sensor-value cpu.power.on
            #sensor-value 1244631611.415231 1 cpu.power.on nominal 0
            !sensor-value ok 1

        """
        exact, name_filter = construct_name_filter(ensure_native_str(msg.arguments[0])
                                                   if msg.arguments else None)
        sensors = [(name, sensor) for name, sensor in
                   sorted(self._sensors.items()) if name_filter(name)]

        if exact and not sensors:
            return req.make_reply("fail", "Unknown sensor name.")

        katcp_version = self.PROTOCOL_INFO.major
        for name, sensor in sensors:
            timestamp, status, value = sensor.read_formatted(katcp_version)
            req.inform(timestamp, "1", name, status, value)
        return req.make_reply("ok", str(len(sensors)))

    def request_sensor_sampling(self, req, msg):
        r"""Configure or query the way a sensor is sampled.

        Sampled values are reported asynchronously using the #sensor-status
        message.

        Parameters
        ----------
        names : str
            One or more names of sensors whose sampling strategy will be queried
            or configured. If specifying multiple sensors, these must be provided as a
            comma-separated list.
            A query can only be done on a single sensor.  However, configuration
            can be done on many sensors with a single request, as long as they
            all use the same strategy.
            Note:  prior to KATCP v5.1 only a single sensor could be configured.
            Multiple sensors are only allowed if the device server sets the protocol
            version to KATCP v5.1 or higher and enables the BULK_SET_SENSOR_SAMPLING
            flag in its PROTOCOL_INFO class attribute.
        strategy : {'none', 'auto', 'event', 'differential', 'differential-rate',
                    'period', 'event-rate'}, optional
            Type of strategy to use to report the sensor value. The
            differential strategy types may only be used with integer or float
            sensors. If this parameter is supplied, it sets the new strategy.
        params : list of str, optional
            Additional strategy parameters (dependent on the strategy type).
            For the differential strategy, the parameter is an integer or float
            giving the amount by which the sensor value may change before an
            updated value is sent.
            For the period strategy, the parameter is the sampling period
            in float seconds.
            The event strategy has no parameters. Note that this has changed
            from KATCPv4.
            For the event-rate strategy, a minimum period between updates and
            a maximum period between updates (both in float seconds) must be
            given. If the event occurs more than once within the minimum period,
            only one update will occur. Whether or not the event occurs, the
            sensor value will be updated at least once per maximum period.
            For the differential-rate strategy there are 3 parameters. The first is the
            same as the differential strategy parameter. The second and third are the
            minimum and maximum periods, respectively, as with the event-rate strategy.

        Informs
        -------
        timestamp : float
            Timestamp of the sensor reading in seconds since the Unix
            epoch, or milliseconds for katcp versions <= 4.
        count : {1}
            Number of sensors described in this #sensor-status inform. Will
            always be one. It exists to keep this inform compatible with
            #sensor-value.
        name : str
            Name of the sensor whose value is being reported.
        value : object
            Value of the named sensor. Type depends on the type of the sensor.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether the sensor-sampling request succeeded.
        names : str
            Name(s) of the sensor queried or configured. If multiple sensors, this will
            be a comma-separated list.
        strategy : {'none', 'auto', 'event', 'differential', 'differential-rate',
                    'period', 'event-rate'}.
            Name of the new or current sampling strategy for the sensor(s).
        params : list of str
            Additional strategy parameters (see description under Parameters).
        Examples
        --------
        ::

            ?sensor-sampling cpu.power.on
            !sensor-sampling ok cpu.power.on none

            ?sensor-sampling cpu.power.on period 0.5
            #sensor-status 1244631611.415231 1 cpu.power.on nominal 1
            !sensor-sampling ok cpu.power.on period 0.5

            if BULK_SET_SENSOR_SAMPLING is enabled then:

            ?sensor-sampling cpu.power.on,fan.speed
            !sensor-sampling fail Cannot\_query\_multiple\_sensors

            ?sensor-sampling cpu.power.on,fan.speed period 0.5
            #sensor-status 1244631611.415231 1 cpu.power.on nominal 1
            #sensor-status 1244631611.415200 1 fan.speed nominal 10.0
            !sensor-sampling ok cpu.power.on,fan.speed period 0.5

        """
        if self._is_bulk_sampling_required(msg.arguments):
            handler = self._handle_bulk_set_sensor_sampling
        else:
            handler = self._handle_sensor_sampling
        f = Future()
        self.ioloop.add_callback(lambda: chain_future(
            handler(req, msg), f))
        return f

    def _is_bulk_sampling_required(self, msg_args):
        bulk_sampling_enabled = self.PROTOCOL_INFO.bulk_set_sensor_sampling
        multiple_sensors_provided = b',' in msg_args[0] if msg_args else False
        strategy_provided = len(msg_args) > 1
        return (
            bulk_sampling_enabled and (multiple_sensors_provided or strategy_provided)
        )

    @gen.coroutine
    def _handle_sensor_sampling(self, req, msg):
        if not msg.arguments:
            raise FailReply("No sensor name given.")

        name = ensure_native_str(msg.arguments[0])

        if name not in self._sensors:
            raise FailReply("Unknown sensor name: %s." % name)

        sensor = self._sensors[name]
        # The client connection that is not specific to this request context
        client = req.client_connection
        katcp_version = self.PROTOCOL_INFO.major

        if len(msg.arguments) > 1:
            # attempt to set sampling strategy
            strategy = msg.arguments[1]
            params = msg.arguments[2:]

            if strategy not in SampleStrategy.SAMPLING_LOOKUP_REV:
                raise FailReply("Unknown strategy name: %s."
                                % ensure_native_str(strategy))

            if not self.PROTOCOL_INFO.strategy_allowed(strategy):
                raise FailReply("Strategy %s not allowed for version %d of katcp"
                                % (ensure_native_str(strategy), katcp_version))

            format_inform = (format_inform_v5
                             if katcp_version >= SEC_TS_KATCP_MAJOR
                             else format_inform_v4)

            def inform_callback(sensor, reading):
                """Inform callback for sensor strategy."""
                timestamp, status, value = reading
                cb_msg = format_inform(sensor, timestamp, status, value)
                client.inform(cb_msg)

            if katcp_version < SEC_TS_KATCP_MAJOR and strategy == b'period':
                # Slightly nasty hack, but since period is the only v4 strategy
                # involving timestamps it's not _too_ nasty :)
                params = [float(params[0]) * MS_TO_SEC_FAC] + params[1:]
            new_strategy = SampleStrategy.get_strategy(
                strategy, inform_callback, sensor, *params, ioloop=self.ioloop)

            # Remove and cancel old strategy
            old_strategy = self._strategies[client].pop(sensor, None)
            if old_strategy:
                old_strategy.cancel()

            # TODO: replace isinstance check with something better
            if not isinstance(new_strategy, SampleNone):
                self._strategies[client][sensor] = new_strategy
                new_strategy.start()

        current_strategy = self._strategies[client].get(sensor, None)
        if not current_strategy:
            current_strategy = SampleStrategy.get_strategy(
                "none", lambda *args: None, sensor)

        strategy, params = current_strategy.get_sampling_formatted()
        if katcp_version < SEC_TS_KATCP_MAJOR and strategy == b'period':
            # Another slightly nasty hack, but since period is the only
            # v4 strategy involving timestamps it's not _too_ nasty :)
            params = [int(float(params[0]) * SEC_TO_MS_FAC)] + params[1:]

        # Let the ioloop run so that the #sensor-status inform is sent before
        # the reply. Not strictly necessary, but a number of tests depend on
        # this behaviour, less effort to fix it here :-/
        yield gen.moment
        raise gen.Return(req.make_reply("ok", name, strategy, *params))

    @gen.coroutine
    def _handle_bulk_set_sensor_sampling(self, req, msg):
        if not msg.arguments:
            raise FailReply("No sensor name given.")

        # The client connection that is not specific to this request context
        client = req.client_connection

        def inform_callback(sensor, reading):
            """Inform callback for sensor strategy."""
            timestamp, status, value = reading
            cb_msg = format_inform_v5(sensor, timestamp, status, value)
            client.inform(cb_msg)

        names_arg = ensure_native_str(msg.arguments[0])
        names = names_arg.split(',')

        if len(msg.arguments) <= 1:
            raise FailReply("Cannot query multiple sensors.")

        strategy = msg.arguments[1]
        params = msg.arguments[2:]

        invalid_names = []
        for name in names:
            if name not in self._sensors:
                invalid_names.append(name)
        if invalid_names:
            raise FailReply("Unknown sensor names: %r." % invalid_names)

        requested_strategies = {}
        try:
            for name in names:
                sensor = self._sensors[name]
                if self.PROTOCOL_INFO.strategy_allowed(strategy):
                    requested_strategy = SampleStrategy.get_strategy(
                        strategy, inform_callback, sensor, *params, ioloop=self.ioloop)
                    requested_strategies[name] = requested_strategy
        except Exception as e:
            raise FailReply("Invalid strategy requested: %s." % e)
        if not requested_strategies:
            raise FailReply(
                "Invalid request: names %r, strategy %r, params %r." % (
                    names,
                    ensure_native_str(strategy),
                    [ensure_native_str(p) for p in params]
                )
            )

        for name in names:
            sensor = self._sensors[name]

            # Remove and cancel old strategy
            old_strategy = self._strategies[client].pop(sensor, None)
            if old_strategy:
                old_strategy.cancel()

            # Replace with new strategy
            new_strategy = requested_strategies[name]
            if not isinstance(new_strategy, SampleNone):
                self._strategies[client][sensor] = new_strategy
                new_strategy.start()

        # Let the ioloop run so that the #sensor-status informs are sent before
        # the reply. Not strictly neccesary, but a number of tests depend on
        # this behaviour, less effort to fix it here :-/
        yield gen.moment

        strategy, params = requested_strategy.get_sampling_formatted()
        reply_args = [",".join(names)] + [strategy] + params
        raise gen.Return(req.make_reply("ok", *reply_args))


    @request()
    @return_reply()
    def request_sensor_sampling_clear(self, req):
        """Set all sampling strategies for this client to none.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether sending the list of devices succeeded.

        Examples
        --------
        ?sensor-sampling-clear
        !sensor-sampling-clear ok

        """
        f = Future()
        @gen.coroutine
        def _clear_strategies():
            self.clear_strategies(req.client_connection)
            raise gen.Return(('ok',))

        self.ioloop.add_callback(lambda: chain_future(_clear_strategies(), f))
        return f

    def request_watchdog(self, req, msg):
        """Check that the server is still alive.

        Returns
        -------
            success : {'ok'}

        Examples
        --------
        ::

            ?watchdog
            !watchdog ok

        """
        # not a function, just doesn't use self
        # pylint: disable-msg = R0201
        return req.make_reply("ok")

    # pylint: enable-msg = W0613


class AsyncDeviceServer(DeviceServer):
    """`DeviceServer` that is automatically configured for async use.

    Same as instantiating a :class:`DeviceServer` instance and calling methods
    `set_concurrency_options(thread_safe=False, handler_thread=False)` and
    `set_ioloop(tornado.ioloop.IOLoop.current())` before starting.

    """
    def __init__(self, *args, **kwargs):
        super(AsyncDeviceServer, self).__init__(*args, **kwargs)
        self.set_concurrency_options(thread_safe=False, handler_thread=False)
        self.set_ioloop(tornado.ioloop.IOLoop.current())

class DeviceLogger(object):
    """Object for logging messages from a DeviceServer.

    Log messages are logged at a particular level and under
    a particular name. Names use dotted notation to form
    a virtual hierarchy of loggers with the device.

    Parameters
    ----------
    device_server : DeviceServerBase object
        The device server this logger should use for sending out logs.
    root_logger : str
        The name of the root logger.

    """

    # level values are used as indexes into the LEVELS list
    # so these to lists should be in the same order
    ALL, TRACE, DEBUG, INFO, WARN, ERROR, FATAL, OFF = range(8)

    ## @brief List of logging level names.
    LEVELS = ["all", "trace", "debug", "info", "warn",
              "error", "fatal", "off"]

    ## @brief Map of Python logging level to corresponding to KATCP levels
    PYTHON_LEVEL = {
        TRACE: 0,
        DEBUG: logging.DEBUG,
        INFO: logging.INFO,
        WARN: logging.WARN,
        ERROR: logging.ERROR,
        FATAL: logging.FATAL,
        OFF: logging.FATAL + 10 # OFF is the highest possible logging level
    }

    def __init__(self, device_server, root_logger="root", python_logger=None):
        self._device_server = device_server
        self._python_logger = python_logger
        self._log_level = self.WARN
        self._root_logger_name = root_logger

    def level_name(self, level=None):
        """Return the name of the given level value.

        If level is None, return the name of the current level.

        Parameters
        ----------
        level : logging level constant
            The logging level constant whose name to retrieve.

        Returns
        -------
        level_name : str
            The name of the logging level.

        """
        if level is None:
            level = self._log_level
        return self.LEVELS[level]

    def level_from_name(self, level_name):
        """Return the level constant for a given name.

        If the *level_name* is not known, raise a ValueError.

        Parameters
        ----------
        level_name : str or bytes
            The logging level name whose logging level constant
            to retrieve.

        Returns
        -------
        level : logging level constant
            The logging level constant associated with the name.

        """
        try:
            level_name = ensure_native_str(level_name)
            return self.LEVELS.index(level_name)
        except ValueError:
            raise ValueError("Unknown logging level name '%s'" % (level_name,))

    def set_log_level(self, level):
        """Set the logging level.

        Parameters
        ----------
        level : logging level constant
            The value to set the logging level to.

        """
        self._log_level = level
        if self._python_logger:
            try:
                level = self.PYTHON_LEVEL.get(level)
            except ValueError:
                raise FailReply("Unknown logging level '%s'" % (level))
            self._python_logger.setLevel(level)

    def set_log_level_by_name(self, level_name):
        """Set the logging level using a level name.

        Parameters
        ----------
        level_name : str or bytes
            The name of the logging level.

        """
        self.set_log_level(self.level_from_name(level_name))

    def log(self, level, msg, *args, **kwargs):
        """Log a message and inform all clients.

        Parameters
        ----------
        level : logging level constant
            The level to log the message at.
        msg : str
            The text format for the log message.
        args : list of objects
            Arguments to pass to log format string. Final message text is
            created using: msg % args.
        kwargs : additional keyword parameters
            Allowed keywords are 'name' and 'timestamp'. The name is the name
            of the logger to log the message to. If not given the name defaults
            to the root logger. The timestamp is a float in seconds. If not
            given the timestamp defaults to the current time.

        """
        timestamp = kwargs.get("timestamp")
        python_msg = msg
        if self._python_logger is not None:
            if timestamp is not None:
                python_msg = ' '.join((
                    'katcp timestamp: %r' % timestamp,
                    python_msg))
            self._python_logger.log(self.PYTHON_LEVEL[level], python_msg, *args)
        if level >= self._log_level:
            name = kwargs.get("name")
            if name is None:
                name = self._root_logger_name
            try:
                inform_msg = msg % args
            except TypeError:
                # Catch the "not enough arguments for format string" exception.
                inform_msg = "{} {}".format(
                    msg,
                    args if args else '').strip()

            self._device_server.mass_inform(
                self._device_server.create_log_inform(
                    self.level_name(level),
                    inform_msg,
                    name,
                    timestamp=timestamp))

    def trace(self, msg, *args, **kwargs):
        """Log a trace message."""
        self.log(self.TRACE, msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        """Log a debug message."""
        self.log(self.DEBUG, msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        """Log an info message."""
        self.log(self.INFO, msg, *args, **kwargs)

    def warn(self, msg, *args, **kwargs):
        """Log an warning message."""
        self.log(self.WARN, msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """Log an error message."""
        self.log(self.ERROR, msg, *args, **kwargs)

    def fatal(self, msg, *args, **kwargs):
        """Log a fatal error message."""
        self.log(self.FATAL, msg, *args, **kwargs)

    @classmethod
    def log_to_python(cls, logger, msg):
        """Log a KATCP logging message to a Python logger.

        Parameters
        ----------
        logger : logging.Logger object
            The Python logger to log the given message to.
        msg : Message object
            The #log message to create a log entry from.

        """
        (level, timestamp, name, message) = tuple(msg.arguments)
        log_string = "%s %s: %s" % (timestamp, name, message)
        logger.log({b"trace": 0,
                    b"debug": logging.DEBUG,
                    b"info": logging.INFO,
                    b"warn": logging.WARN,
                    b"error": logging.ERROR,
                    b"fatal": logging.FATAL}[level], log_string)



# Copyright 2010 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

""" This is a benchmark client for scenario 1, which cooperate with
benchtxserver or benchserver
"""
from __future__ import absolute_import, division, print_function

import sys
import time
from optparse import OptionParser

from katcp.tx.core import ClientKatCP, run_client
from twisted.internet import reactor
from twisted.python import log
from util import standard_parser

TIMEOUT = 0.2

class DemoClient(ClientKatCP):
    counter = 0
    no_of_sensors = 0 # number of sensors sampled
    last = 0
    sampling = False

    def __init__(self, *args, **kwds):
        ClientKatCP.__init__(self, *args, **kwds)
        self.avg = []

    def inform_sensor_status(self, msg):
        self.last = int(msg.arguments[0])
        self.counter += 1

    def start_sampling(self, _):
        name = 'int_sensor%d' % self.no_of_sensors
        self.no_of_sensors += 1
        self.send_request('sensor-sampling', name, 'period', 1)
        self.sampling = False

    def sample_next_sensor(self):
        if not self.sampling:
            self.sampling = True
            self.send_request('sensor-list').addCallback(self.check_sensor_list)

    def check_sensor_list(self, informs_reply):
        informs, reply = informs_reply
        sensor_no = len(informs)
        if self.no_of_sensors < sensor_no:
            self.start_sampling(None)
        elif self.options.allow_sensor_creation:
            self.send_request('add-sensor').addCallback(self.start_sampling)

    def periodic_check(self):
        self.avg.append(self.counter)
        if len(self.avg) > 10:
            self.avg.pop(0)
        print("AVG: %d, LAST: %d, SENSORS: %d" % (
            sum(self.avg)/len(self.avg), self.counter, self.no_of_sensors))
        sys.stdout.flush()
        if (not self.options.allow_sensor_creation or
            (abs(self.counter - self.no_of_sensors * 200) <=
            (self.no_of_sensors * 100))):
            self.sample_next_sensor()
        self.counter = 0
        reactor.callLater(TIMEOUT, self.periodic_check)

    def connectionLost(self, failure):
        print("Connection lost, exiting", file=sys.stderr)
        if reactor.running:
            reactor.stop()

def connected(protocol, options):
    protocol.options = options
    reactor.callLater(TIMEOUT, protocol.periodic_check)
    protocol.sample_next_sensor()

def not_connected(failure):
    print(failure, file=sys.stderr)
    print("Exiting", file=sys.stderr)
    reactor.stop()

if __name__ == '__main__':
    parser = standard_parser()
    parser.add_option('--allow-sensor-creation', dest='allow_sensor_creation',
                      default=False, action='store_true')
    options, args = parser.parse_args()
    run_client(('localhost', options.port), DemoClient, connected,
               (options,), not_connected)
    reactor.run()



# Copyright 2010 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

""" This is a benchmark device server using twisted katcp impl
"""
from __future__ import absolute_import, division, print_function

import sys

from katcp import Message, Sensor
from katcp.tx import DeviceProtocol, DeviceServer
from katcp.tx.test.testserver import FloatSensor, IntSensor
from twisted.internet import reactor
from twisted.internet.protocol import Factory
from twisted.python import log
from util import standard_parser


class BenchmarkProtocol(DeviceProtocol):
    def request_add_sensor(self, msg):
        self.factory.add_sensor(Sensor(int,
                                   'int_sensor%d' % len(self.factory.sensors),
                                    'descr', 'unit', params=[-10, 10]))
        return Message.reply('add-sensor', 'ok')


class BenchmarkServer(DeviceServer):
    protocol = BenchmarkProtocol

    def setup_sensors(self):
        pass

def main():
    parser = standard_parser()
    options, args = parser.parse_args()
    factory = BenchmarkServer(options.port, '')
    print(factory.start().getHost())
    sys.stdout.flush()
    reactor.run() # run the main twisted reactor

if __name__ == '__main__':
    main()



# Copyright 2010 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

""" This is a benchmark client for scenario 2, which cooperate with
benchtxserver or benchserver
"""
from __future__ import absolute_import, division, print_function

import sys
import time
from optparse import OptionParser

from katcp.tx.core import ClientKatCP, run_client
from twisted.internet import reactor
from twisted.python import log
from util import standard_parser

TIMEOUT = 0.2

class DemoClient(ClientKatCP):
    counter = 0

    def got_sensor_value(self, v):
        self.counter += 1
        self.send_request('sensor-value', 'int_sensor').addCallback(
            self.got_sensor_value)

    def periodic_check(self):
        print(self.counter)
        sys.stdout.flush()
        reactor.callLater(TIMEOUT, self.periodic_check)
        self.counter = 0

def connected(protocol, options):
    protocol.send_request('sensor-value', 'int_sensor').addCallback(
        protocol.got_sensor_value)
    reactor.callLater(TIMEOUT, protocol.periodic_check)

def not_connected(failure):
    print(failure, file=sys.stderr)
    print("Exiting", file=sys.stderr)
    reactor.stop()

if __name__ == '__main__':
    parser = standard_parser()
    options, args = parser.parse_args()
    run_client(('localhost', options.port), DemoClient, connected,
               (options,), not_connected)
    reactor.run()



# Copyright 2010 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

from __future__ import absolute_import, division, print_function

from optparse import OptionParser


def standard_parser(default_port=1235):
    parser = OptionParser()
    parser.add_option('--port', dest='port', type=int, default=default_port)
    return parser



# Copyright 2012 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

from __future__ import absolute_import, division, print_function
from future import standard_library

standard_library.install_aliases()  # noqa: E402

import threading
import time
import unittest

import tornado.gen
import tornado.testing

from katcp import Sensor, testutils


def get_sensor(sensor_type, name=None):
    if name is None:
        name = 'test_%s_sensor' % sensor_type
    params = None
    if sensor_type in (Sensor.INTEGER, Sensor.FLOAT):
        params = [0, 1000]
    elif sensor_type == Sensor.DISCRETE:
        params = ['value1', 'value2', 'value3']
    sensor = Sensor(
        sensor_type, name, "Dummy %s Sensor" % sensor_type, "Units",
        params)

    return sensor


class test_SensorTransitionWaiter(unittest.TestCase):
    def test_wait_float_timeout(self):
        timeout = 0.1
        expected_conditions = (
            lambda x: x < 0.7,
            lambda x: x >= 0.7,
            lambda x: x >= 1,
            lambda x: x < 1,
            lambda x: x < 0.3)
        value_series = (0.5, 0.6, 1, 0.7, 0.299)

        def sensor_stream():  # Target for thread that pushes values to DUT
            # We are purposefully not adjusting the timestamps (1st 2
            # parameters to sensor.set_value()) to check that events
            # that come in faster than timer precision are also caught
            thread_alive.set()
            for val in value_series:
                time.sleep(delay_value)
                sensor.set_value(val)

        # Set up the DUT
        sensor = get_sensor(Sensor.FLOAT)
        DUT = testutils.SensorTransitionWaiter(sensor, expected_conditions)

        # Set up the thread that will push the values to DUT
        sensor_thread = threading.Thread(target=sensor_stream)
        sensor_thread.daemon = True
        thread_alive = threading.Event()
        delay_value = 0.005  # Should be fast enough to complete within timeout
        thread_alive.clear()
        sensor_thread.start()
        # wait until the thread starts working
        thread_alive.wait(timeout=0.5)
        self.assertTrue(DUT.wait(timeout=timeout))
        self.assertFalse(DUT.timed_out)
        sensor_thread.join()

        # Now try it too slow
        delay_value = 0.021
        DUT = testutils.SensorTransitionWaiter(sensor, expected_conditions)
        sensor_thread = threading.Thread(target=sensor_stream)
        sensor_thread.daemon = True
        thread_alive.clear()
        # wait until the thread starts working
        sensor_thread.start()
        thread_alive.wait(timeout=0.5)
        self.assertFalse(DUT.wait(timeout=timeout))
        self.assertTrue(DUT.timed_out)
        sensor_thread.join()

    def test_init_teardown(self):
        now = time.time()
        sensor = get_sensor(Sensor.INTEGER)
        sensor.set_value(0)
        # Test that an assertion is raised if the initial value of the
        # sensor does not match the first value in the expected
        # sequence.
        with self.assertRaises(ValueError):
            testutils.SensorTransitionWaiter(sensor, (1, 2, 3))

        DUT = testutils.SensorTransitionWaiter(sensor, (0, 2, 3))
        # Test that we are attached to the sensor
        self.assertTrue(DUT._observer in sensor._observers)
        self.assertEqual(DUT._observer.update, DUT._sensor_callback)

        DUT.teardown()
        # Check that the callback is unregistered
        self.assertFalse(DUT._observer in sensor._observers)
        self.assertTrue(DUT._torn_down)
        with self.assertRaises(RuntimeError):
            DUT.wait()  # should not allow waiting if we're torn down


class test_wait_sensor(unittest.TestCase):
    def _wait_sensor(self, vals, val, status):
        sensor = get_sensor(Sensor.INTEGER)
        sensor.set_value(0, Sensor.NOMINAL)

        def set_vals():
            time.sleep(0.05)
            for v in vals:
                if status is None:
                    sensor.set_value(v)
                else:
                    sensor.set_value(*v)

        sensor_thread = threading.Thread(target=set_vals)
        # test timeout
        self.assertFalse(testutils.wait_sensor(sensor, val, status=status, timeout=0.05))
        # Now start setting vals
        sensor_thread.start()
        self.assertTrue(testutils.wait_sensor(sensor, val, status=status, timeout=1))

    def test_values(self):
        vals = (1, 2, 0, 3)
        self._wait_sensor(vals, 3, status=None)

    def test_values_and_status(self):
        vals = ((1, Sensor.NOMINAL),
                (7, Sensor.WARN),
                (2, Sensor.ERROR),
                (0, Sensor.ERROR))
        self._wait_sensor(vals, 0, status=Sensor.ERROR)


class _test_WaitingMockBase(unittest.TestCase):
    DUTClass = None

    def test_reset_mock(self):
        # skip test run on base clase directly
        if not self.DUTClass:
            return
        # Verify that the call_count and call_args_list variables
        # are initially zero, and get cleared by calling reset_mock
        DUT = self.DUTClass()
        self.assertEqual(DUT.call_count, 0)
        self.assertEqual(len(DUT.call_args_list), 0)
        DUT()
        self.assertEqual(DUT.call_count, 1)
        self.assertEqual(len(DUT.call_args_list), 1)
        DUT.reset_mock()
        self.assertEqual(DUT.call_count, 0)
        self.assertEqual(len(DUT.call_args_list), 0)


class test_WaitingMock(_test_WaitingMockBase):
    DUTClass = testutils.WaitingMock

    def test_assert_wait_call_count_success(self):
        # Test the normal case, in which the mock was called
        DUT = self.DUTClass()
        DUT(123)
        DUT.assert_wait_call_count(1, timeout=0.1)

    def test_assert_wait_call_count_fail_on_call_count(self):
        # Test the negative case, when the mock was not called.
        # This should cause an assertion error after the timeout.
        DUT = self.DUTClass()
        with self.assertRaises(AssertionError):
            DUT.assert_wait_call_count(1, timeout=0.01)

    def test_assert_wait_call_count_fail_on_call_args(self):
        # Synthetic test case where the call_count is correct, but the
        # call_args_list has not been updated yet.  This might occur if
        # the mock call is done in a different thread to the unit test.
        # In this case we expect a runtime error after the timeout.
        DUT = self.DUTClass()
        DUT(123)
        DUT.call_args_list = []
        with self.assertRaises(RuntimeError):
            DUT.assert_wait_call_count(1, timeout=0.1)


class test_AsyncWaitingMock(tornado.testing.AsyncTestCase, _test_WaitingMockBase):
    DUTClass = testutils.AsyncWaitingMock

    @tornado.testing.gen_test
    def test_assert_wait_call_count_success(self):
        # Test the normal case, in which the mock was called
        DUT = self.DUTClass()
        DUT(123)
        yield DUT.assert_wait_call_count(1, timeout=0.1)

    @tornado.testing.gen_test
    def test_assert_wait_call_count_fail_on_call_count(self):
        # Test the negative case, when the mock was not called.
        # This should cause an assertion error after the timeout.
        DUT = self.DUTClass()
        with self.assertRaises(AssertionError):
            yield DUT.assert_wait_call_count(1, timeout=0.01)

    @tornado.testing.gen_test
    def test_ioloop_hogging(self):
        # Test implementation detail, if the waiting loop forgets to clear
        # DUT._call_event after each call, it will go in a "busy waiting" loop
        # that blocks the ioloop and prevents the second call (done using
        # add_callback) from happening until after assert_wait_call_count()
        # times out.
        DUT = self.DUTClass()
        DUT(1)
        self.io_loop.add_callback(DUT, 2)
        yield DUT.assert_wait_call_count(2, timeout=0.1)



# Copyright 2017 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details
from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

import sys
import unittest

from builtins import object
from concurrent.futures import Future

from _thread import get_ident as get_thread_ident

# Module under test
from katcp import ioloop_manager
from katcp.testutils import start_thread_with_cleanup


class test_ThreadsafeMethodAttrWrapper(unittest.TestCase):
    def setUp(self):
        self.ioloop_manager = ioloop_manager.IOLoopManager(managed_default=True)
        self.ioloop = self.ioloop_manager.get_ioloop()
        self.ioloop_thread_wrapper = ioloop_manager.IOLoopThreadWrapper(self.ioloop)
        start_thread_with_cleanup(self, self.ioloop_manager, start_timeout=1)

    def test_wrapping(self):
        test_inst = self
        class Wrappee(object):
            def __init__(self, ioloop_thread_id):
                self.thread_id = ioloop_thread_id

            def a_callable(self, arg, kwarg='abc'):
                test_inst.assertEqual(get_thread_ident(), self.thread_id)
                return (arg * 2, kwarg * 3)

            @property
            def not_in_ioloop(self):
                test_inst.assertNotEqual(get_thread_ident(), self.thread_id)
                return 'not_in'

            @property
            def only_in_ioloop(self):
                test_inst.assertEqual(get_thread_ident(), self.thread_id)
                return 'only_in'

        class TestWrapper(ioloop_manager.ThreadSafeMethodAttrWrapper):
            @property
            def only_in_ioloop(self):
                return self._getattr('only_in_ioloop')


        id_future = Future()
        self.ioloop.add_callback(lambda : id_future.set_result(get_thread_ident()))
        wrappee = Wrappee(id_future.result(timeout=1))
        wrapped = TestWrapper(wrappee, self.ioloop_thread_wrapper)
        # First test our assumptions about Wrappee
        with self.assertRaises(AssertionError):
            wrappee.a_callable(3, 'a')
        with self.assertRaises(AssertionError):
            wrappee.only_in_ioloop
        self.assertEqual(wrappee.not_in_ioloop, 'not_in')

        # Now test the wrapped version
        self.assertEqual(wrapped.a_callable(5, kwarg='bcd'), (10, 'bcd'*3))
        self.assertEqual(wrapped.only_in_ioloop, 'only_in')
        self.assertEqual(wrapped.not_in_ioloop, 'not_in')


class test_IOLoopManager(unittest.TestCase):
    def setUp(self):
        self.ioloop_manager = ioloop_manager.IOLoopManager(managed_default=True)
        self.ioloop = self.ioloop_manager.get_ioloop()
        start_thread_with_cleanup(self, self.ioloop_manager, start_timeout=1)

    def test_managed_io_loop_is_asyncio_in_python3(self):
        if sys.version_info[0] == 3:
            self.assertTrue(hasattr(self.ioloop, "asyncio_loop"))
        else:
            self.assertFalse(hasattr(self.ioloop, "asyncio_loop"))



# Copyright 2014 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details
"""A high-level abstract interface to KATCP clients, sensors and requests."""

from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

import abc
import collections
import logging
import sys

from builtins import object

import tornado

from future.utils import with_metaclass, PY2
from past.builtins import basestring
from tornado.concurrent import Future
from tornado.gen import Return, with_timeout

from katcp import Message, Sensor
from katcp.core import hashable_identity

logger = logging.getLogger(__name__)


class KATCPResourceError(Exception):
    """Error raised for resource-related errors"""


class KATCPResourceInactive(KATCPResourceError):
    """Raised when a request is made to an inactive resource"""


class KATCPSensorError(KATCPResourceError):
    """Raised if a problem occurred dealing with as KATCPSensor operation"""


class SensorResultTuple(collections.namedtuple(
        'SensorResultTuple',
        'object name python_identifier description type units reading')):
    """Per-sensor result of list_sensors() method

    Attributes
    ----------
    object : KATCPSensor instance
    name : str
        KATCP (i.e. unescaped) name of the sensor
    python_identifier : str
        Python-identifier name of the sensor.
    description : str
        KATCP description of the sensor
    type : str
        KATCP type of the sensor
    units : str
        KATCP units of the sensor
    reading : KATCPSensorReading instance
        Most recently received sensor reading
    """
    __slots__ = []  # Prevent dynamic attributes from being possible


def normalize_strategy_parameters(params):
    """Normalize strategy parameters to be a list of strings.

    Parameters
    ----------
    params : (space-delimited) string or sequence of strings/numbers Parameters
        expected by :class:`SampleStrategy` object, in various forms, where the first
        parameter is the name of the strategy.

    Returns
    -------
    params : tuple of strings
        Strategy parameters as a list of strings

    """
    def fixup_numbers(val):
        try:
            # See if it is a number
            return str(float(val))
        except ValueError:
            # ok, it is not a number we know of, perhaps a string
            return str(val)
    if isinstance(params, basestring):
        params = params.split(' ')
    # No number
    return tuple(fixup_numbers(p) for p in params)


def escape_name(name):
    """Escape sensor and request names to be valid Python identifiers."""
    return name.replace('.', '_').replace('-', '_')


class KATCPResource(with_metaclass(abc.ABCMeta, object)):

    """Base class to serve as the definition of the KATCPResource API.

    A class `C` implementing the KATCPResource API should register itself using
    KATCPResource.register(C) or subclass KATCPResource directly. A complication
    involved with subclassing is that all the abstract properties must be
    implemented as properties; normal instance attributes cannot be used.

    Attributes
    ----------
    Apart from the abstract properties described below

    TODO Describe how hierarchies are implemented. Also all other descriptions
    here so that the sphinx doc can be auto-generated from here.

    """

    def __init__(self):
        self._active = True

    @abc.abstractproperty
    def name(self):
        """Name of this KATCP resource."""

    @abc.abstractproperty
    def description(self):
        """Description of this KATCP resource."""

    @abc.abstractproperty
    def address(self):
        """Address of the underlying client/device.

        Type: tuple(host, port) or None, with host a string and port an integer.

        If this KATCPResource is not associated with a specific KATCP device
        (e.g. it is only a top-level container for a hierarchy of KATCP
        resources), the address should be None.

        """

    @abc.abstractproperty
    def is_connected(self):
        """Indicate whether the underlying client/device is connected or not."""

    @abc.abstractproperty
    def req(self):
        """Attribute root/container for all KATCP request wrappers.

        Each KATCP request that is exposed on a KATCP device should have a
        corresponding :class:`KATCPRequest` object so that calling

          `resource.req.request_name(arg1, arg2, ...)`

        sends a '?request-name arg1 arg2 ...' message to the KATCP device and
        waits for the associated inform-reply and reply messages.

        For a :class:`KATCPResource` object that exposes a hierarchical device
        it can choose to include lower-level request handlers here such that
        `resource.req.dev_request()` maps to `resource.dev.req.request()`.

        """

    @abc.abstractproperty
    def sensor(self):
        """Attribute root/container for all KATCP sensor wrappers.

        Each KATCP sensor that is exposed on a KATCP device should have a
        corresponding :class:`KATCPSensor` object so that

          `resource.sensor.sensor_name`

        corresponds to a sensor named e.g. 'sensor-name', where the object or
        attribute name is an escaped/Pythonised version of the original sensor
        name (see :func:`escape_name` for the escape mechanism). Hopefully the
        device is not crazy enough to have multiple sensors that map to the
        same Python identifier.

        A :class:`KATCPResource` object that exposes a hierarchical device can
        choose to include lower-level sensors here such that
        `resource.sensor.dev_sensorname` maps to
        `resource.dev.sensor.sensorname`.

        """

    @abc.abstractproperty
    def parent(self):
        """Parent KATCPResource object of this subordinate resource, or None."""

    @abc.abstractproperty
    def children(self):
        """AttrDict of subordinate KATCPResource objects keyed by their names."""

    @tornado.gen.coroutine
    def wait(self, sensor_name, condition_or_value, timeout=5):
        """Wait for a sensor in this resource to satisfy a condition.

        Parameters
        ----------
        sensor_name : string
            The name of the sensor to check
        condition_or_value : obj or callable, or seq of objs or callables
            If obj, sensor.value is compared with obj. If callable,
            condition_or_value(reading) is called, and must return True if its
            condition is satisfied. Since the reading is passed in, the value,
            status, timestamp or received_timestamp attributes can all be used
            in the check.
        timeout : float or None
            The timeout in seconds (None means wait forever)

        Returns
        -------
        This command returns a tornado Future that resolves with True when the
        sensor value satisfies the condition, or False if the condition is
        still not satisfied after a given timeout period.

        Raises
        ------
        :class:`KATCPSensorError`
            If the sensor does not have a strategy set, or if the named sensor
            is not present

        """
        sensor_name = escape_name(sensor_name)
        sensor = self.sensor[sensor_name]
        try:
            yield sensor.wait(condition_or_value, timeout)
        except tornado.gen.TimeoutError:
            raise tornado.gen.Return(False)
        else:
            raise tornado.gen.Return(True)

    @abc.abstractmethod
    def list_sensors(self, filter="", strategy=False, status="",
                     use_python_identifiers=True, tuple=False, refresh=False):
        """List sensors available on this resource matching certain criteria.

        Parameters
        ----------
        filter : string, optional
            Filter each returned sensor's name against this regexp if specified.
            To ease the dichotomy between Python identifier names and actual
            sensor names, the default is to search on Python identifier names
            rather than KATCP sensor names, unless `use_python_identifiers`
            below is set to False. Note that the sensors of subordinate
            KATCPResource instances may have inconsistent names and Python
            identifiers, better to always search on Python identifiers in this
            case.
        strategy : {False, True}, optional
            Only list sensors with a set strategy if True
        status : string, optional
            Filter each returned sensor's status against this regexp if given
        use_python_identifiers : {True, False}, optional
            Match on python identifiers even the the KATCP name is available.
        tuple : {True, False}, optional, Default: False
            Return backwards compatible tuple instead of SensorResultTuples
        refresh : {True, False}, optional, Default: False
            If set the sensor values will be refreshed with get_value before
            returning the results.

        Returns
        -------
        sensors : list of SensorResultTuples, or list of tuples
            List of matching sensors presented as named tuples. The `object`
            field is the :class:`KATCPSensor` object associated with the sensor.
            Note that the name of the object may not match `name` if it
            originates from a subordinate device.
        """

    @tornado.gen.coroutine
    def set_sampling_strategies(self, filter, strategy_and_params):
        """Set a sampling strategy for all sensors that match the specified filter.

        Parameters
        ----------
        filter : string
            The regular expression filter to use to select the sensors to which
            to apply the specified strategy.  Use "" to match all sensors. Is
            matched using :meth:`list_sensors`.
        strategy_and_params : seq of str or str
            As tuple contains (<strat_name>, [<strat_parm1>, ...]) where the strategy
            names and parameters are as defined by the KATCP spec. As str contains the
            same elements in space-separated form.
        **list_sensor_args : keyword arguments
            Passed to the :meth:`list_sensors` call as kwargs

        Returns
        -------
        sensors_strategies : tornado Future
           resolves with a dict with the Python identifier names of the sensors
           as keys and the value a tuple:

           (success, info) with

           success : bool
              True if setting succeeded for this sensor, else False
           info : tuple
               normalised sensor strategy and parameters as tuple if success == True
               else, sys.exc_info() tuple for the error that occurred.
        """
        sensors_strategies = {}
        sensor_results = yield self.list_sensors(filter)
        for sensor_reslt in sensor_results:
            norm_name = sensor_reslt.object.normalised_name
            try:
                sensor_strat = yield self.set_sampling_strategy(norm_name, strategy_and_params)
                sensors_strategies[norm_name] = sensor_strat[norm_name]
            except Exception:
                sensors_strategies[norm_name] = (
                    False, sys.exc_info())
        raise tornado.gen.Return(sensors_strategies)

    @tornado.gen.coroutine
    def set_sampling_strategy(self, sensor_name, strategy_and_params):
        """Set a sampling strategy for a specific sensor.

        Parameters
        ----------
        sensor_name : string
            The specific sensor.
        strategy_and_params : seq of str or str
            As tuple contains (<strat_name>, [<strat_parm1>, ...]) where the strategy
            names and parameters are as defined by the KATCP spec. As str contains the
            same elements in space-separated form.

        Returns
        -------
        sensors_strategies : tornado Future
           resolves with a dict with the Python identifier names of the sensors
           as keys and the value a tuple:

           (success, info) with

           success : bool
              True if setting succeeded for this sensor, else False
           info : tuple
               normalised sensor strategy and parameters as tuple if success == True
               else, sys.exc_info() tuple for the error that occurred.
        """
        sensors_strategies = {}
        try:
            sensor_obj = self.sensor.get(sensor_name)
            yield sensor_obj.set_sampling_strategy(strategy_and_params)
            sensors_strategies[sensor_obj.normalised_name] = (
                    True, sensor_obj.sampling_strategy)
        except Exception:
            sensors_strategies[sensor_obj.normalised_name] = (
                    False, sys.exc_info())
        raise tornado.gen.Return(sensors_strategies)

    def set_active(self, active):
        self._active = bool(active)
        for child in dict.values(self.children):
            child.set_active(active)

    def is_active(self):
        return self._active


class KATCPRequest(with_metaclass(abc.ABCMeta, object)):

    """Abstract Base class to serve as the definition of the KATCPRequest API.

    Wrapper around a specific KATCP request to a given KATCP device.  Each
    available KATCP request for a particular device has an associated
    :class:`KATCPRequest` object in the object hierarchy. This wrapper is mainly
    for interactive convenience. It provides the KATCP request help string as a
    docstring and pretty-prints the result of the request.

    """

    def __init__(self, request_description, is_active=lambda: True):
        """Initialize request with given description and network client

        Parameters
        ----------
        request_description : dict
           name : str
            KATCP name of the request
           description : str
            KATCP request description (as returned by ?help <name>)
           timeout_hint : float or None
            Request timeout suggested by device or None if not provided
        is_active : callable, optional
            Returns True if this request is active, else False

        """
        for required_description_key in ('name', 'description', 'timeout_hint'):
            if required_description_key not in request_description:
                raise ValueError(
                    'Required request_description key {!r} not present'
                    .format(required_description_key))
        self._request_description = dict(request_description)
        self.__doc__ = '\n'.join(('KATCP Documentation',
                                  '===================',
                                  self.description,
                                  'KATCPRequest Documentation',
                                  '==========================',
                                  self.__doc__ or ''))
        self._is_active = is_active

    @property
    def name(self):
        """Name of the KATCP request."""
        return self._request_description['name']

    @property
    def description(self):
        """Description of KATCP request as obtained from the ?help request."""
        return self._request_description['description']

    @property
    def timeout_hint(self):
        """Request timeout suggested by device or None if not provided"""
        return self._request_description['timeout_hint']


    def __call__(self, *args, **kwargs):
        """Execute the KATCP request described by this object.

        All positional arguments of this function are converted to KATCP string
        representations and passed on as space-separated parameters to the KATCP
        device.

        Keyword Arguments
        -----------------
        timeout : None or float, optional
            Timeout in seconds for the request. If None, use request timeout
            hint received from server or default for the :class:`KATCPResource`
            instance that contains the request if no hint is available.
        mid : None or int, optional
            Message identifier to use for the request message. If None, use
            either auto-incrementing value or no mid depending on the KATCP
            protocol version (mid's were only introduced with KATCP v5) and the
            default of the containing :class:`KATCPResource` instance.

        Returns
        -------
        reply : tornado future resolving with :class:`KATCPReply` object
            KATCP request reply wrapped in KATCPReply object

        Raises
        ------
        :class:`ResourceInactive` if the resource is inactive when the request is made.
        """
        if self.is_active():
            return self.issue_request(*args, **kwargs)
        else:
            raise KATCPResourceInactive(
                "Can't make ?{} request; resource is inactive".format(self.name))

    @abc.abstractmethod
    def issue_request(self, *args, **kwargs):
        """Signature as for __call__

        Do the request immediately without checking active state.
        """

    def is_active(self):
        """True if resource for this request is active"""
        return self._is_active()

class KATCPDummyRequest(KATCPRequest):
    """Dummy counterpart to KATCPRequest that always returns a successful reply"""
    def issue_request(self, *args, **kwargs):
        reply_msg = Message.reply('fake', 'ok')
        reply = KATCPReply(reply_msg, [])
        fut = Future()
        fut.set_result(reply)
        return fut

class KATCPSensorReading(collections.namedtuple(
        'KATCPSensorReading', 'received_timestamp timestamp istatus value')):

    """Sensor reading as a (received_timestamp, timestamp, istatus, value) tuple.

    Attributes
    ----------
    received_timestamp : float
       Time (in seconds since UTC epoch) at which the sensor value was received.
    timestamp : float
       Time (in seconds since UTC epoch) at which the sensor value was determined.
    istatus : int Sensor status constant
       Whether the value represents an error condition or not, as in class:`katcp.Sensor`
       The status is stored as an int, but output as a string, eg 'nominal'.
    value : object
        The value of the sensor (the type will be appropriate to the
        sensor's type).
    """

    __slots__ = []              # Prevent dynamic attributes

    @property
    def status(self):
        " Returns the string representation of sensor status, eg 'nominal'"
        try:
            return Sensor.STATUSES[int(self.istatus)]
        except TypeError:
            return 'unknown'


class KATCPSensorsManager(with_metaclass(abc.ABCMeta, object)):

    """Sensor management class used by KATCPSensor. Abstracts communications details.

    This class should arrange:

    1. A mechanism for setting sensor strategies
    2. A mechanism for polling a sensor value
    3. Keeping track of- and reapplying sensor strategies after reconnect, etc.
    4. Providing local time. This is doing to avoid direct calls to time.time, allowing
       accelerated time testing / simulation / dry-running
    """

    @abc.abstractmethod
    def time(self):
        """Returns the current time (in seconds since UTC epoch)"""

    @abc.abstractmethod
    def get_sampling_strategy(self, sensor_name):
        """Get the current sampling strategy for the named sensor

        Parameters
        ----------

        sensor_name : str
            Name of the sensor (normal or escaped form)

        Returns
        -------

        strategy : tornado Future that resolves with tuple of str
            contains (<strat_name>, [<strat_parm1>, ...]) where the strategy names and
            parameters are as defined by the KATCP spec
        """

    @abc.abstractmethod
    def set_sampling_strategy(self, sensor_name, strategy_and_parms):
        """Set the sampling strategy for the named sensor

        Parameters
        ----------

        sensor_name : str
            Name of the sensor
        strategy : seq of str or str
            As tuple contains (<strat_name>, [<strat_parm1>, ...]) where the strategy
            names and parameters are as defined by the KATCP spec. As str contains the
            same elements in space-separated form.

        Returns
        -------

        done : tornado Future that resolves when done or raises KATCPSensorError

        Notes
        -----

        It is recommended that implementations use :func:`normalize_strategy_parameters`
        to process the strategy_and_parms parameter, since it will deal with both string
        and list versions and makes sure that numbers are represented as strings in a
        consistent format.

        This method should arrange for the strategy to be set on the underlying network
        device or whatever other implementation is used. This strategy should also be
        automatically re-set if the device is reconnected, etc. If a strategy is set for a
        non-existing sensor, it should still cache the strategy and ensure that is applied
        whenever said sensor comes into existence. This allows an applications to pre-set
        strategies for sensors before synced / connected to a device.

        """

    @abc.abstractmethod
    def drop_sampling_strategy(self, sensor_name):
        """Drop the sampling strategy for the named sensor from the cache

        Calling :meth:`set_sampling_strategy` requires the sensor manager to
        memorise the requested strategy so that it can automatically be reapplied.
        If the client is no longer interested in the sensor, or knows the sensor
        may be removed from the server, then it can use this method to ensure the
        manager forgets about the strategy.  This method will not change the current
        strategy.  No error is raised if there is no strategy to drop.

        Parameters
        ----------

        sensor_name : str
            Name of the sensor (normal or escaped form)

        """

    @abc.abstractmethod
    def poll_sensor(self, sensor_name):
        """Poll sensor and arrange for sensor object to be updated

        Returns
        -------

        done_future : tornado Future
            Resolves when the poll is complete, or raises KATCPSensorError
        """
        # TODO NM 2015-02-03 Might want to add a timeout parameter here, and to all the
        # other code that calls this

    @abc.abstractmethod
    def reapply_sampling_strategies(self):
        """Reapply all sensor strategies using cached values

        Would typically be called when a connection is re-established. Should
        not raise errors when resetting strategies for sensors that no longer
        exist on the KATCP resource.
        """


class KATCPSensor(with_metaclass(abc.ABCMeta, object)):
    """Wrapper around a specific KATCP sensor on a given KATCP device.

    Each available KATCP sensor for a particular device has an associated
    :class:`KATCPSensor` object in the object hierarchy. This wrapper is mainly
    for interactive convenience. It provides the KATCP request help string as a
    docstring and registers listeners. Subclasses need to call the base class
    version of __init__().
    """

    def __init__(self, sensor_description, sensor_manager):
        """Subclasses must arrange to call this in their __init__().

        Parameters
        ----------
        sensor_description : dict
           Description of the KATCP sensor, with keys same as the parameters of
           :class:`katcp.Sensor`
        sensor_manager : :class:`KATCPSensorsManager` instance
           Manages sensor strategies, allows sensor polling, and provides time
        """
        self._manager = sensor_manager
        self.clear_listeners()
        self._reading = KATCPSensorReading(0, 0, Sensor.UNKNOWN, None)
        # We'll be abusing a katcp.Sensor object slightly to make use of its
        # parsing and formatting functionality
        self._sensor = Sensor(**sensor_description)
        self._name = self._sensor.name
        # Overide the katpc.Sensor's set method with ours
        self._sensor.set = self.set
        # Steal the the katcp.Sensor's set_formatted method. Since we overrode
        # its set() method with ours, calling set_formatted will result in this
        # KATCPSensor object's value being set.
        self.set_formatted = self._sensor.set_formatted

    @property
    def parent_name(self):
        """Name of the parent of this KATCPSensor"""
        return self._manager.resource_name

    @property
    def name(self):
        """Name of this KATCPSensor"""
        return self._name

    @property
    def normalised_name(self):
        """Normalised name of this KATCPSensor that can be used as a python identifier"""
        return escape_name(self._name)

    @property
    def reading(self):
        """Most recently received sensor reading as KATCPSensorReading instance"""
        return self._reading

    @property
    def value(self):
        return self._reading.value

    @property
    def status(self):
        return self._reading.status

    @property
    def sampling_strategy(self):
        """Current sampling strategy"""
        return self._manager.get_sampling_strategy(self.name)

    @property
    def description(self):
        return self._sensor.description

    @property
    def units(self):
        return self._sensor.units

    @property
    def type(self):
        return self._sensor.type

    def parse_value(self, s_value):
        """Parse a value from a string.

        Parameters
        ----------
        s_value : str
            A string value to attempt to convert to a value for
            the sensor.

        Returns
        -------
        value : object
            A value of a type appropriate to the sensor.

        """
        return self._sensor.parse_value(s_value)

    def set_strategy(self, strategy, params=None):
        """Set current sampling strategy for sensor.
        Add this footprint for backwards compatibility.

        Parameters
        ----------

        strategy : seq of str or str
            As tuple contains (<strat_name>, [<strat_parm1>, ...]) where the strategy
            names and parameters are as defined by the KATCP spec. As str contains the
            same elements in space-separated form.
        params : seq of str or str
            (<strat_name>, [<strat_parm1>, ...])

        Returns
        -------
        done : tornado Future that resolves when done or raises KATCPSensorError

        """
        if not params:
            param_args = []
        elif isinstance(params, basestring):
            param_args = [str(p) for p in params.split(' ')]
        else:
            if not isinstance(params, collections.Iterable):
                params = (params,)
            param_args = [str(p) for p in params]
        samp_strategy = " ".join([strategy] + param_args)
        return self._manager.set_sampling_strategy(self.name, samp_strategy)

    def set_sampling_strategy(self, strategy):
        """Set current sampling strategy for sensor

        Parameters
        ----------

        strategy : seq of str or str
            As tuple contains (<strat_name>, [<strat_parm1>, ...]) where the strategy
            names and parameters are as defined by the KATCP spec. As str contains the
            same elements in space-separated form.

        Returns
        -------
        done : tornado Future that resolves when done or raises KATCPSensorError

        """
        return self._manager.set_sampling_strategy(self.name, strategy)

    def drop_sampling_strategy(self):
        """Drop memorised sampling strategy for sensor, if any

        Calling this method ensures that the sensor manager does not attempt
        to reapply a sampling strategy.  It will not raise an error if no strategy
        has been set.  Use :meth:`set_sampling_strategy` to memorise a strategy again.
        """
        self._manager.drop_sampling_strategy(self.name)

    def register_listener(self, listener, reading=False):
        """Add a callback function that is called when sensor value is updated.

        The callback footprint is received_timestamp, timestamp, status, value.

        Parameters
        ----------
        listener : function
            Callback signature: if reading listener(katcp_sensor, reading) where
            `katcp_sensor` is this KATCPSensor instance `reading` is an instance of
            :class:`KATCPSensorReading`.

            Callback signature: default, if not reading listener(received_timestamp,
            timestamp, status, value)
        """
        listener_id = hashable_identity(listener)
        self._listeners[listener_id] = (listener, reading)
        logger.debug(
                    'Register listener for {}'
                    .format(self.name))

    def unregister_listener(self, listener):
        """Remove a listener callback added with register_listener().

        Parameters
        ----------
        listener : function
            Reference to the callback function that should be removed

        """
        listener_id = hashable_identity(listener)
        self._listeners.pop(listener_id, None)

    def is_listener(self, listener):
        listener_id = hashable_identity(listener)
        return listener_id in self._listeners

    def clear_listeners(self):
        """Clear any registered listeners to updates from this sensor."""
        self._listeners = {}

    def call_listeners(self, reading):
        logger.debug(
                    'Calling listeners {}'
                    .format(self.name))
        for listener, use_reading in list(self._listeners.values()):
            try:
                if use_reading:
                    listener(self, reading)
                else:
                    listener(reading.received_timestamp, reading.timestamp,
                             reading.status, reading.value)
            except Exception:
                logger.exception(
                    'Unhandled exception calling KATCPSensor callback {0!r}'
                    .format(listener))

    def set(self, timestamp, status, value):
        """Set sensor with a given received value, matches :meth:`katcp.Sensor.set`"""
        received_timestamp = self._manager.time()
        reading = KATCPSensorReading(received_timestamp, timestamp, status, value)
        self._reading = reading
        self.call_listeners(reading)

    def set_value(self, value, status=Sensor.NOMINAL, timestamp=None):
        """Set sensor value with optinal specification of status and timestamp"""
        if timestamp is None:
            timestamp = self._manager.time()
        self.set(timestamp, status, value)

    def set_formatted(self, raw_timestamp, raw_status, raw_value, major):
        """Set sensor using KATCP string formatted inputs

        Mirrors :meth:`katcp.Sensor.set_formatted`.

        This implementation is empty. Will, during instantiation, be overridden by the
        set_formatted() method of a katcp.Sensor object.
        """

    @tornado.gen.coroutine
    def get_reading(self):
        """Get a fresh sensor reading from the KATCP resource

        Returns
        -------
        reply : tornado Future resolving with  :class:`KATCPSensorReading` object

        Notes
        -----

        As a side-effect this will update the reading stored in this object, and result in
        registered listeners being called.
        """
        yield self._manager.poll_sensor(self._name)
        # By now the sensor manager should have set the reading
        raise Return(self._reading)

    @tornado.gen.coroutine
    def get_value(self):
        """Get a fresh sensor value from the KATCP resource

        Returns
        -------
        reply : tornado Future resolving with  :class:`KATCPSensorReading` object

        Notes
        -----

        As a side-effect this will update the reading stored in this object, and result in
        registered listeners being called.
        """
        yield self._manager.poll_sensor(self._name)
        # By now the sensor manager should have set the reading
        raise Return(self._reading.value)

    @tornado.gen.coroutine
    def get_status(self):
        """Get a fresh sensor status from the KATCP resource

        Returns
        -------
        reply : tornado Future resolving with  :class:`KATCPSensorReading` object

        Notes
        -----

        As a side-effect this will update the reading stored in this object, and result in
        registered listeners being called.
        """
        yield self._manager.poll_sensor(self._name)
        # By now the sensor manager should have set the reading
        raise Return(self._reading.status)

    def wait(self, condition_or_value, timeout=None):
        """Wait for the sensor to satisfy a condition.

        Parameters
        ----------
        condition_or_value : obj or callable, or seq of objs or callables
            If obj, sensor.value is compared with obj. If callable,
            condition_or_value(reading) is called, and must return True if its
            condition is satisfied. Since the reading is passed in, the value,
            status, timestamp or received_timestamp attributes can all be used
            in the check.
            TODO: Sequences of conditions (use SensorTransitionWaiter thingum?)
        timeout : float or None
            The timeout in seconds (None means wait forever)

        Returns
        -------
        This command returns a tornado Future that resolves with True when the
        sensor value satisfies the condition. It will never resolve with False;
        if a timeout is given a TimeoutError happens instead.

        Raises
        ------
        :class:`KATCPSensorError`
            If the sensor does not have a strategy set
        :class:`tornado.gen.TimeoutError`
            If the sensor condition still fails after a stated timeout period

        """
        if (isinstance(condition_or_value, collections.Sequence) and not
                isinstance(condition_or_value, basestring)):
            raise NotImplementedError(
                'Currently only single conditions are supported')
        condition_test = (condition_or_value if callable(condition_or_value)
                          else lambda s: s.value == condition_or_value)

        ioloop = tornado.ioloop.IOLoop.current()
        f = Future()
        if self.sampling_strategy == ('none', ):
            raise KATCPSensorError(
                'Cannot wait on a sensor that does not have a strategy set')

        def handle_update(sensor, reading):
            # This handler is called whenever a sensor update is received
            try:
                assert sensor is self
                if condition_test(reading):
                    self.unregister_listener(handle_update)
                    # Try and be idempotent if called multiple times after the
                    # condition is matched. This should not happen unless the
                    # sensor object is being updated in a thread outside of the
                    # ioloop.
                    if not f.done():
                        ioloop.add_callback(f.set_result, True)
            except Exception:
                f.set_exc_info(sys.exc_info())
                self.unregister_listener(handle_update)

        self.register_listener(handle_update, reading=True)
        # Handle case where sensor is already at the desired value
        ioloop.add_callback(handle_update, self, self._reading)

        if timeout:
            to = ioloop.time() + timeout
            timeout_f = with_timeout(to, f)
            # Make sure we stop listening if the wait times out to prevent a
            # buildup of listeners
            timeout_f.add_done_callback(
                lambda f: self.unregister_listener(handle_update))
            return timeout_f
        else:
            return f


_KATCPReplyTuple = collections.namedtuple('_KATCPReplyTuple', 'reply informs')


class KATCPReply(_KATCPReplyTuple):
    """Container for return messages of KATCP request (reply and informs).

    This is based on a named tuple with 'reply' and 'informs' fields so that
    the :class:`KATCPReply` object can still be unpacked into a normal tuple.

    Parameters
    ----------
    reply : :class:`katcp.Message` object
        Reply message returned by katcp request
    informs : list of :class:`katcp.Message` objects
        List of inform messages returned by KATCP request

    Attributes
    ----------
    messages: list of :class:`katcp.Message` objects
        List of all messages returned by KATCP request, reply first
    reply:  :class:`katcp.Message` object
        Reply message returned by KATCP request
    informs: list of :class:`katcp.Message` objects
        List of inform messages returned by KATCP request

    The instance evaluates to nonzero (i.e. truthy) if the request succeeded.

    """

    __slots__ = []  # Prevent dynamic attributes from being possible

    def __repr__(self):
        """String representation for pretty-printing in IPython."""
        return '\n'.join(
            "%s%s %s" %
            (Message.TYPE_SYMBOLS[m.mtype], m.name, ' '.join(m.arguments))
            for m in self.messages)

    def __str__(self):
        """String representation using KATCP wire format"""
        return '\n'.join(str(m) for m in self.messages)

    def __bool__(self):
        """True if request succeeded (i.e. first reply argument is 'ok')."""
        return self.messages[0].reply_ok()

    if PY2:
        __nonzero__ = __bool__

    @property
    def messages(self):
        """List of all messages returned by KATCP request, reply first."""
        return [self.reply] + self.informs

    @property
    def succeeded(self):
        """True if request succeeded (i.e. first reply argument is 'ok')."""
        return bool(self)



# Copyright 2014 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

from __future__ import absolute_import, division, print_function

import logging

import mock

from katcp.server import *



# Copyright 2015 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

import logging
import unittest

import mock
import tornado.testing

# Module under test
from katcp import Sensor, resource
from katcp.testutils import TimewarpAsyncTestCase


logger = logging.getLogger(__name__)


class test_escape_name(unittest.TestCase):
    def test_escape_name(self):
        desired_mappings = {
            'blah-bal' : 'blah_bal',
            'bief_bof.ba32f-blief' : 'bief_bof_ba32f_blief',
            'already_escape_name' : 'already_escape_name'}
        for input, expected_output in desired_mappings.items():
            self.assertEqual(resource.escape_name(input), expected_output)

class test_KATCPSensor(TimewarpAsyncTestCase):
    @tornado.testing.gen_test
    def test_wait(self):
        sensor_manager = mock.Mock()
        sensor_manager.get_sampling_strategy.return_value = (
            resource.normalize_strategy_parameters('none'))
        DUT = resource.KATCPSensor(dict(sensor_type=Sensor.INTEGER,
                                        name='test.int', params=[-1000, 1000],
                                        default=0, initial_status=Sensor.NOMINAL),
                                   sensor_manager)
        # Test that an exception is raised if no strategy is set
        with self.assertRaises(resource.KATCPSensorError):
            yield DUT.wait(1)
        # Check that no stray listeners are left behind
        self.assertFalse(DUT._listeners)

        # Test that timeout works
        sensor_manager.get_sampling_strategy.return_value = (
            resource.normalize_strategy_parameters('event'))
        # Add an ioloop callback to timewarp so that we don't get stuck
        timeout = 0.25
        self.io_loop.add_callback(
            self.set_ioloop_time, self.io_loop.time() + timeout + 0.1)
        # Check that the timout error is raised
        with self.assertRaises(tornado.gen.TimeoutError):
            yield DUT.wait(1, timeout=timeout)
        # Check that no stray listeners are left behind
        self.assertFalse(DUT._listeners)

        # Wait for a value that changes within the timeout
        self.io_loop.add_callback(
            self.set_ioloop_time, self.io_loop.time() + timeout - 0.1)
        waiting_value = 2
        self.io_loop.add_callback(DUT.set_value, waiting_value)
        yield DUT.wait(waiting_value, timeout=timeout)
        self.assertEqual(DUT.value, waiting_value)
        # Check that no stray listeners are left behind
        self.assertFalse(DUT._listeners)

        # Wait using a callable, comparison times out
        self.io_loop.add_callback(
            self.set_ioloop_time, self.io_loop.time() + timeout + 0.1)
        waiting_value = 11
        waiting_condition = lambda reading: reading.value == waiting_value
        # Check that the timout error is raised
        with self.assertRaises(tornado.gen.TimeoutError):
            yield DUT.wait(waiting_value, timeout=timeout)
        # Check that no stray listeners are left behind
        self.assertFalse(DUT._listeners)

        # Wait for a callable condition that is reached within the timeout
        self.io_loop.add_callback(
            self.set_ioloop_time, self.io_loop.time() + timeout - 0.1)
        waiting_value = 12
        waiting_condition = lambda reading: reading.value == waiting_value
        self.io_loop.add_callback(DUT.set_value, waiting_value)
        yield DUT.wait(waiting_condition, timeout=timeout)
        self.assertEqual(DUT.value, waiting_value)
        # Check that no stray listeners are left behind
        self.assertFalse(DUT._listeners)

        # wait for value and status
        waiting_value = 3
        waiting_status = 'warn'
        waiting_condition = lambda reading: reading.value == waiting_value and \
                                            reading.status == waiting_status
        self.io_loop.add_callback(DUT.set_value, 3, Sensor.WARN)
        yield DUT.wait(waiting_condition, timeout=timeout)
        self.assertEqual(DUT.value, waiting_value)
        self.assertEqual(DUT.status, waiting_status)
        # Check that no stray listeners are left behind
        self.assertFalse(DUT._listeners)

    @tornado.testing.gen_test
    def test_sampling_strategy_property(self):
        sensor_name = 'test.int'
        strategy = 'testing 123'
        sensor_manager = mock.Mock()
        sensor_manager.get_sampling_strategy.return_value = strategy
        DUT = resource.KATCPSensor(dict(sensor_type=Sensor.INTEGER,
                                        name=sensor_name),
                                   sensor_manager)
        self.assertEqual(DUT.sampling_strategy, strategy)

    @tornado.testing.gen_test
    def test_set_strategy(self):
        sensor_name = 'test.int'
        sensor_manager = mock.Mock()
        DUT = resource.KATCPSensor(dict(sensor_type=Sensor.INTEGER,
                                        name=sensor_name),
                                   sensor_manager)
        DUT.set_strategy('none')
        DUT.set_strategy('period', 0.5)
        DUT.set_strategy('event-rate', '1.1 1.2')
        DUT.set_strategy('event-rate', [2.1, 2.2])

        sensor_manager.set_sampling_strategy.assert_has_calls([
            mock.call(sensor_name, 'none'),
            mock.call(sensor_name, 'period 0.5'),
            mock.call(sensor_name, 'event-rate 1.1 1.2'),
            mock.call(sensor_name, 'event-rate 2.1 2.2'),
        ])

    @tornado.testing.gen_test
    def test_set_sampling_strategy(self):
        sensor_name = 'test.int'
        strategy = 'period 2.5'
        sensor_manager = mock.Mock()
        DUT = resource.KATCPSensor(dict(sensor_type=Sensor.INTEGER,
                                        name=sensor_name),
                                   sensor_manager)
        DUT.set_sampling_strategy(strategy)
        sensor_manager.set_sampling_strategy.assert_called_once_with(
            sensor_name, strategy)

    @tornado.testing.gen_test
    def test_drop_sampling_strategy(self):
        sensor_name = 'test.int'
        sensor_manager = mock.Mock()
        DUT = resource.KATCPSensor(dict(sensor_type=Sensor.INTEGER,
                                        name=sensor_name),
                                   sensor_manager)
        DUT.drop_sampling_strategy()
        sensor_manager.drop_sampling_strategy.assert_called_once_with(
            sensor_name)


class ConcreteKATCPRequest(resource.KATCPRequest):
    def issue_request(self, *args, **kwargs):
        pass

class test_KATCPRequest(unittest.TestCase):
    def test_active(self):
        active = False
        is_active = lambda: active
        req_name = 'test-request'
        req_description = {
            'name': req_name,
            'description': '?test-request description',
            'timeout_hint': None}

        DUT = ConcreteKATCPRequest(req_description, is_active=is_active)
        DUT.issue_request = mock.Mock()

        req_args = ('arg1', 'arg2')
        req_kwargs = dict(timeout=123, mid=456)

        # Do a test request when active = False, should raise
        with self.assertRaises(resource.KATCPResourceInactive):
            DUT(*req_args, **req_kwargs)
        self.assertEqual(
            DUT.issue_request.call_count, 0,
            'issue_request should not have been called when request is inactive')

        # now set active to True, and check that the request is made
        active = True
        rv = DUT(*req_args, **req_kwargs)
        self.assertEqual(rv, DUT.issue_request.return_value)
        DUT.issue_request.assert_called_once_with(*req_args, **req_kwargs)



# Copyright 2014 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

from __future__ import absolute_import, division, print_function

import logging
import signal
import threading
import time

import IPython
import tornado

from katcp import inspecting_client, resource_client
from katcp.testutils import DeviceTestServer

logging.basicConfig(
    format="%(asctime)s %(name)s %(levelname)s %(funcName)s(%(filename)s:%(lineno)d)%(message)s",
    level=logging.DEBUG
)

log = logging.getLogger(__name__)

ioloop = tornado.ioloop.IOLoop.current()

d = DeviceTestServer('', 0)
d.set_concurrency_options(False, False)
d.set_ioloop(ioloop)
ioloop.add_callback(d.start)

def setup_resource_client():
    global rc
    print(d.bind_address)
    rc = resource_client.KATCPClientResource(dict(
        name='thething',
        address=d.bind_address,
        controlled=True
    ))
    rc.start()

def printy(*args):
    print(args)

@tornado.gen.coroutine
def setup_inspecting_client():
    global ic
    try:
        print(d.bind_address)
        host, port = d.bind_address
        ic = inspecting_client.InspectingClientAsync(host, port, ioloop=ioloop)
        ic.set_state_callback(printy)
        yield ic.connect()
    except Exception:
        log.exception('whups')

#ioloop.add_callback(setup_inspecting_client)
ioloop.add_callback(setup_resource_client)

stop = threading.Event()

@tornado.gen.coroutine
def doreq(req, *args, **kwargs):
    print('hi')
    try:
        rep = yield req(*args, **kwargs)
        print(rep)
    except Exception:
        print('logging')
        log.exception('oops')
    finally:
        print('blah')

def run_ipy():
    try:
        iotw = resource_client.IOLoopThreadWrapper(ioloop)
        time.sleep(0.24)
        s = rc.sensor.an_int
        ws = resource_client.ThreadsafeKATCPSensorWrapper(s, iotw)
        IPython.embed()
        # stop.wait(10000)
    finally:
        ioloop.stop()

t = threading.Thread(target=run_ipy)
t.start()

ioloop.set_blocking_log_threshold(0.1)
def ignore_signal(sig, frame):
    pass
# Disable sigint, since it stops the ioloop but not ipython shell ;)
signal.signal(signal.SIGINT, ignore_signal)
try:
    ioloop.start()
except KeyboardInterrupt:
    print('Keyboard interrupt')
    stop.set()
